"""
This is a module that aims to list all
possibilities of data formats offered by Dessia.
It should be used as a repertory of rules and available typings.

Some general rules :

- Lists are homogeneous, ie, they should not
  contain several types of elements
    ex : List[int], List[str], List[CustomClass], ...
- Tuples can, therefore, be used as heterogenous sequences,
  thanks to the fact that they are immutables.
    ex : Tuple[str, int, CustomClass] is a tuple like this :
        t = ('tuple', 1, custom_class_object)
- Dict is used whenever a dynamic structure should be defined.
  It takes only one possible type for keys
  and one possible type for values.
    ex : Dict[str, bool] is a dict like :
        d = {'key0': True, 'key1': False, 'another_key': False,...}
- As opposed to this, a non-standalone_in_db class defines a static structure,
  with a defined number of given, expected keys & types of their values.

In addition to types & genericity (brought by DessiaObject),
this module can also be seen as a template for Dessia's
coding/naming style & convention.
"""

from math import floor, ceil
from typing import Dict, List, Tuple, Union, TextIO, BinaryIO

try:
    import volmdlr as vm
    from volmdlr.wires import Contour2D
    from volmdlr import primitives2d as p2d
    from volmdlr import primitives3d as p3d
    import plot_data
    from plot_data.colors import *
except:
    pass

from dessia_common import DessiaObject
from dessia_common.typings import InstanceOf, Distance
from dessia_common.vectored_objects import Catalog

from numpy import linspace
from math import cos

from dessia_common.files import BinaryFile, StringFile
import io


class StandaloneSubobject(DessiaObject):
    _standalone_in_db = True
    _generic_eq = True

    def __init__(self, floatarg: Distance, name: str = 'Standalone Subobject'):
        self.floatarg = floatarg

        DessiaObject.__init__(self, name=name)

    @classmethod
    def generate(cls, seed: int) -> 'StandaloneSubobject':
        floatarg = Distance(1.7 * seed)
        name = 'StandaloneSubobject' + str(seed)
        return cls(floatarg=floatarg, name=name)

    @classmethod
    def generate_many(cls, seed: int) -> List['StandaloneSubobject']:
        subobjects = [cls.generate((i + 1) * 1000) for i in range(seed)]
        return subobjects

    def contour(self):
        points = [vm.Point2D(0, 0), vm.Point2D(0, 1),
                  vm.Point2D(1, 1), vm.Point2D(1, 0)]

        crls = p2d.ClosedRoundedLineSegments2D(points=points, radius={})
        return vm.wires.Contour2D(crls.primitives)

    def voldmlr_primitives(self):
        contour = self.contour()
        volumes = [p3d.ExtrudedProfile(vm.O3D, vm.X3D, vm.Z3D,
                                       contour, [], vm.Y3D)]
        return volumes


DEF_SS = StandaloneSubobject.generate(1)


class EnhancedStandaloneSubobject(StandaloneSubobject):
    def __init__(self, floatarg: Distance, boolarg: bool,
                 name: str = 'Standalone Subobject'):
        self.boolarg = boolarg

        StandaloneSubobject.__init__(self, floatarg=floatarg, name=name)

    @classmethod
    def generate(cls, seed: int) -> 'EnhancedStandaloneSubobject':
        floatarg = Distance(1.2 * seed)
        boolarg = floatarg.is_integer()
        name = 'EnhancedStandaloneSubobject' + str(seed)
        return cls(floatarg=floatarg, boolarg=boolarg, name=name)


DEF_ESS = EnhancedStandaloneSubobject.generate(1)


class InheritingStandaloneSubobject(StandaloneSubobject):
    def __init__(self, floatarg: Distance, strarg: str,
                 name: str = 'Inheriting Standalone Subobject'):
        self.strarg = strarg

        StandaloneSubobject.__init__(self, floatarg=floatarg, name=name)

    @classmethod
    def generate(cls, seed: int) -> 'InheritingStandaloneSubobject':
        floatarg = Distance(0.7 * seed)
        strarg = str(-seed)
        name = 'Inheriting Standalone Subobject' + str(seed)
        return cls(floatarg=floatarg, strarg=strarg, name=name)


DEF_ISS = InheritingStandaloneSubobject.generate(1)


class EmbeddedSubobject(DessiaObject):
    def __init__(self, embedded_list: List[int] = None,
                 name: str = 'Embedded Subobject'):
        if embedded_list is None:
            self.embedded_list = [1, 2, 3]
        else:
            self.embedded_list = embedded_list

        DessiaObject.__init__(self, name=name)

    @classmethod
    def generate(cls, seed: int) -> 'EmbeddedSubobject':
        if not bool(seed % 2):
            embedded_list = list(range(int(seed / 2)))
        else:
            embedded_list = None
        name = 'Embedded Subobject' + str(seed)
        return cls(embedded_list=embedded_list, name=name)

    @classmethod
    def generate_many(cls, seed: int) -> List['EmbeddedSubobject']:
        return [cls.generate(i) for i in range(ceil(seed / 3))]


class EnhancedEmbeddedSubobject(EmbeddedSubobject):
    def __init__(self, embedded_list: List[int] = None,
                 embedded_array: List[List[float]] = None,
                 name: str = 'Enhanced Embedded Subobject'):
        self.embedded_array = embedded_array

        EmbeddedSubobject.__init__(self, embedded_list=embedded_list,
                                   name=name)

    @classmethod
    def generate(cls, seed: int) -> 'EnhancedEmbeddedSubobject':
        embedded_list = [seed]
        embedded_array = [[seed, seed * 10, seed * 10]] * seed
        name = 'Embedded Subobject' + str(seed)
        return cls(embedded_list=embedded_list, embedded_array=embedded_array,
                   name=name)


DEF_ES = EmbeddedSubobject.generate(10)
DEF_EES = EnhancedEmbeddedSubobject.generate(3)


UnionArg = Union[EmbeddedSubobject, EnhancedEmbeddedSubobject]


class StandaloneObject(DessiaObject):
    """
    Dev Object for testing purpose

    :param standalone_subobject: A dev subobject that is standalone_in_db
    :type standalone_subobject: StandaloneSubobject
    :param embedded_subobject: A dev subobject that isn't standalone_in_db
    :type embedded_subobject: EmbeddedSubobject
    :param dynamic_dict: A variable length dict
    :type dynamic_dict: Dict[str, bool]
    :param tuple_arg: A heterogeneous sequence
    :type tuple_arg: tuple
    """
    _standalone_in_db = True
    _generic_eq = True
    _allowed_methods = ['add_standalone_object', 'add_embedded_object',
                        'add_float', 'generate_from_text', 'generate_from_bin',
                        'generate_from_bin_file', 'generate_from_text_file']

    def __init__(self, standalone_subobject: StandaloneSubobject,
                 embedded_subobject: EmbeddedSubobject,
                 dynamic_dict: Dict[str, bool],
                 float_dict: Dict[str, float],
                 string_dict: Dict[str, str],
                 tuple_arg: Tuple[str, int],
                 intarg: int, strarg: str,
                 object_list: List[StandaloneSubobject],
                 subobject_list: List[EmbeddedSubobject],
                 builtin_list: List[int],
                 union_arg: List[UnionArg],
                 subclass_arg: InstanceOf[StandaloneSubobject],
                 array_arg: List[List[float]],
                 name: str = 'Standalone Object Demo'):
        self.union_arg = union_arg
        self.builtin_list = builtin_list
        self.subobject_list = subobject_list
        self.object_list = object_list
        self.tuple_arg = tuple_arg
        self.strarg = strarg
        self.intarg = intarg
        self.dynamic_dict = dynamic_dict
        self.float_dict = float_dict
        self.string_dict = string_dict
        self.standalone_subobject = standalone_subobject
        self.embedded_subobject = embedded_subobject
        self.subclass_arg = subclass_arg
        self.array_arg = array_arg

        DessiaObject.__init__(self, name=name)

    @classmethod
    def generate(cls, seed: int,
                 name: str = 'Standalone Object Demo') -> 'StandaloneObject':
        is_even = not bool(seed % 2)
        standalone_subobject = StandaloneSubobject.generate(seed)
        embedded_subobject = EmbeddedSubobject.generate(seed)
        dynamic_dict = {'n' + str(i): bool(seed % 2) for i in range(seed)}
        float_dict = {'k' + str(i): seed * 1.09 for i in range(seed)}
        string_dict = {'key' + str(i): 'value' + str(i) for i in range(seed)}
        tuple_arg = ('value', seed * 3)
        intarg = seed
        strarg = str(seed) * floor(seed / 3)
        object_list = StandaloneSubobject.generate_many(seed)
        subobject_list = EmbeddedSubobject.generate_many(seed)
        builtin_list = [seed] * seed
        array_arg = [builtin_list] * 3
        union_arg = [EnhancedEmbeddedSubobject.generate(seed),
                     EmbeddedSubobject.generate(seed)]
        if is_even:
            subclass_arg = StandaloneSubobject.generate(-seed)
        else:
            subclass_arg = InheritingStandaloneSubobject.generate(seed)
        return cls(standalone_subobject=standalone_subobject,
                   embedded_subobject=embedded_subobject,
                   dynamic_dict=dynamic_dict,
                   float_dict=float_dict,
                   string_dict=string_dict,
                   tuple_arg=tuple_arg,
                   intarg=intarg, strarg=strarg, object_list=object_list,
                   subobject_list=subobject_list, builtin_list=builtin_list,
                   union_arg=union_arg, subclass_arg=subclass_arg,
                   array_arg=array_arg, name=name)

    @classmethod
    def generate_from_text(cls, stream: TextIO):
        try:
            my_string = stream.read()
            # this is a hack for test until we get frontend support for types BinaryFile & StringFile
            # a TextIO does not have filename, but it's ok since we return a StringFile from backend
            my_file_name = stream.filename
            name, raw_seed = my_string.split(",")
            seed = int(raw_seed.strip())
        finally:
            stream.close()
        return cls.generate(seed=seed, name=my_file_name)

    @classmethod
    def generate_from_bin(cls, stream: BinaryIO):
        # the user need to decode the binary as he see fit
        try:
            my_string = stream.read().decode('utf8')
            # this is a hack for test until we get frontend support for types BinaryFile & StringFile
            # a BinaryIO does not have filename, but it's ok since we return a BinaryFile from backend
            my_file_name = stream.filename
            my_name, raw_seed = my_string.split(",")
            seed = int(raw_seed.strip())
        finally:
            stream.close()
        return cls.generate(seed=seed, name=my_file_name)

    @classmethod
    def generate_from_bin_file(cls, stream: BinaryFile):
        # the user need to decode the binary as he see fit
        try:
            my_string = stream.read().decode('utf8')
            my_file_name = stream.filename
            my_name, raw_seed = my_string.split(",")
            seed = int(raw_seed.strip())
        finally:
            stream.close()
        return cls.generate(seed=seed, name=my_file_name)

    @classmethod
    def generate_from_text_file(cls, stream: StringFile):
        try:
            my_text = stream.read()
            my_file_name = stream.filename
            name, raw_seed = my_text.split(",")
            seed = int(raw_seed.strip())
        finally:
            stream.close()
        return cls.generate(seed=seed, name=my_file_name)

    def add_standalone_object(self, object_: StandaloneSubobject):
        """
        This methods adds a standalone object to object_list.
        It doesn't return anything, hence, API will update object when
        computing from frontend
        """
        self.object_list.append(object_)

    def add_embedded_object(self, object_: EmbeddedSubobject):
        """
        This methods adds an embedded object to subobject_list.
        It doesn't return anything, hence, API will update object
        when computing from frontend
        """
        self.subobject_list.append(object_)

    def add_float(self, value: float) -> StandaloneSubobject:
        """
        This methods adds value to its standalone subobject
        floatarg property and returns it.
        API should replace standalone_subobject as it is returned
        """
        self.standalone_subobject.floatarg += value
        return self.standalone_subobject

    def volmdlr_primitives(self):
        return self.standalone_subobject.voldmlr_primitives()

    def plot_data(self):
        attributes = ['cx', 'cy']

        # Contour
        contour = self.standalone_subobject.contour().plot_data()
        primitives_group = plot_data.PrimitiveGroup(primitives=[contour],
                                                    name='Contour')

        # Scatter Plot
        bounds = {'x': [0, 6], 'y': [100, 2000]}
        catalog = Catalog.random_2d(bounds=bounds, threshold=8000)
        points = [plot_data.Point2D(cx=v[0], cy=v[1], name='Point' + str(i))
                  for i, v in enumerate(catalog.array)]
        axis = plot_data.Axis()
        tooltip = plot_data.Tooltip(attributes=attributes,
                                    name='Tooltips')
        scatter_plot = plot_data.Scatter(axis=axis, tooltip=tooltip,
                                         elements=points,
                                         x_variable=attributes[0],
                                         y_variable=attributes[1],
                                         name='Scatter Plot')

        # Parallel Plot
        attributes = ['cx', 'cy', 'color_fill', 'color_stroke']
        parallel_plot = plot_data.ParallelPlot(elements=points,
                                               axes=attributes,
                                               name='Parallel Plot')

        # Multi Plot
        objects = [scatter_plot, parallel_plot]
        sizes = [plot_data.Window(width=560, height=300),
                 plot_data.Window(width=560, height=300)]
        coords = [(0, 0), (300, 0)]
        multi_plot = plot_data.MultiplePlots(elements=points, plots=objects,
                                             sizes=sizes, coords=coords,
                                             name='Multiple Plot')

        attribute_names = ['time', 'electric current']
        tooltip = plot_data.Tooltip(attributes=attribute_names)
        time1 = linspace(0, 20, 20)
        current1 = [t ** 2 for t in time1]
        elements1 = []
        for time, current in zip(time1, current1):
            elements1.append({'time': time, 'electric current': current})

        # The previous line instantiates a dataset with limited arguments but
        # several customizations are available
        point_style = plot_data.PointStyle(color_fill=RED, color_stroke=BLACK)
        edge_style = plot_data.EdgeStyle(color_stroke=BLUE, dashline=[10, 5])

        custom_dataset = plot_data.Dataset(elements=elements1, name='I = f(t)',
                                           tooltip=tooltip,
                                           point_style=point_style,
                                           edge_style=edge_style)

        # Now let's create another dataset for the purpose of this exercice
        time2 = linspace(0, 20, 100)
        current2 = [100 * (1 + cos(t)) for t in time2]
        elements2 = []
        for time, current in zip(time2, current2):
            elements2.append({'time': time, 'electric current': current})

        dataset2 = plot_data.Dataset(elements=elements2, name='I2 = f(t)')

        graph2d = plot_data.Graph2D(graphs=[custom_dataset, dataset2],
                                    x_variable=attribute_names[0],
                                    y_variable=attribute_names[1])
        return [primitives_group, scatter_plot,
                parallel_plot, multi_plot, graph2d]

    def maldefined_method(self, arg0, arg1=1, arg2: int = 10, arg3=3):
        """
        Defining a docstring for testing parsing purpose
        """
        nok_string = "This is a bad coding behavior"
        ok_string = "This could be OK as temporary attr"
        self.maldefined_attr = nok_string
        self._ok_attribute = ok_string

        computation = nok_string + 'or' + ok_string

        return computation

    def to_markdown(self):
        contents = """
        # Quem Stygios dumque

        ## Recursus erat aere decus Lemnicolae

        Lorem markdownum laetum senior quod Libys utroque
        *mirantibus teneat aevo*, aquis.
        Procumbit eandem ensis, erigor intercepta, quae habitabat nostro
        *et hoc que* enim: inpulit.
        Mecum ferat **fecissem** vale per myricae suis
        quas turba potentior mentita.
        Annis nunc, picae erat quis minatur dare Diana redimitus
        [Clymene venisses sinat](http://est.net/umbram.html)
        protinus pulchra, sucos! Tanta haec varios tuaque,
        nisi Erigonen si aquae Hippomene inguine murmur.

        1. Poma enim dextra icta capillis extinctum foedera
        2. Mediis requirit exercita ascendere fecisse sola
        3. Sua externis tigride saevarum
        4. Aves est pendebant sume latentis

        ## Suum videre quondam generis dolentem simul femineos

        Ille lacus progenitore Cycnum pressa, excidit silva
        [crudus](http://www.domino.com/nequevox), boum ducem vocari,
        ne monte tanto harenae.
        Opus Aesone excipit adempto.
        Inpius illa latratu atque sed praedam,
        ille construit intravit concipit,
        concha dedit, qua audit calathosque.
        Dedit putrefacta cortex.
        Tenet aut carmina quod proditione media; pro ense medicina
        vita repetit adrectisque inops e sentiat.

        > Imagine caesaries superbos muneraque *ne terras* cunctis.
        Diversae Hesioneque
        > numinis regia at anima nascuntur Iovis.
        Sua fama quoque capillos lugubris
        > **egimus**, a ingenti [Ericthonio](http://raptos.org/lucem)
        iubebat!

        ## Ponderis venit veteris mihi tofis

        Propensum discedunt, iacere dedisti; lene potest caelo,
        felix flamma caecus decet excipit.
        *Aurum occiderat*, retro cum, quorum *Diana timuere At*.
        Ait Labros hasta mundi, **ut est** ruit nosse o gravet!

        ## Qui aether undis nulla

        Homines oppidaque nominibus devexo genitoris quoque,
        praesensque rota Saturnia.
        Auras cecinit fera quae mirantum imbris,
        Gratia verba incesto, sed visa contigit
        saepe adicit trepidant.
        [Siqua radiis quod](http://www.naris-pectebant.org/comeset)
        ad duabus alienisque, sponte; dum.

        Occidit Babylonia dubitare. Vultus cui: erat dea!
        Iam ense forma est se, tibi pedem adfectat nec nostra.
        Armenta socium nutrix [precatur](http://in-fraxinus.io/)
        aderam, quam mentem Elin labor auctor
        potentia prodidit inmitibus duo di?
        Verum a, tuo quoque nec Mysum per posses;
        vigor danda meruit: tecum audire responsa
        [conplexae](http://quis.io/disrestat.html) et alios.

        Agros grata illo animo mei nova, in magis furens et
        [modo](http://pondere.com/aquis) dimittere ubi neque es!
        Sua qua ac ire una facit Alcmene coepere
        arduus quae vestigia aliquis; meritorum Dorylas, scindunt.
        """
        return contents


DEF_SO = StandaloneObject.generate(1)


class StandaloneObjectWithDefaultValues(StandaloneObject):
    _non_editable_attributes = ['intarg', 'strarg']

    def __init__(self, standalone_subobject: StandaloneSubobject = DEF_SS,
                 embedded_subobject: EmbeddedSubobject = DEF_ES,
                 dynamic_dict: Dict[str, bool] = None,
                 float_dict: Dict[str, float] = None,
                 string_dict: Dict[str, str] = None,
                 tuple_arg: Tuple[str, int] = ("Default Tuple", 0),
                 intarg: int = 1, strarg: str = "Default Strarg",
                 object_list: List[StandaloneSubobject] = None,
                 subobject_list: List[EmbeddedSubobject] = None,
                 builtin_list: List[int] = None,
                 union_arg: List[UnionArg] = None,
                 subclass_arg: InstanceOf[StandaloneSubobject] = DEF_ISS,
                 array_arg: List[List[float]] = None,
                 name: str = 'Standalone Object Demo'):
        if dynamic_dict is None:
            dynamic_dict = {}
        if float_dict is None:
            float_dict = {}
        if string_dict is None:
            string_dict = {}
        if object_list is None:
            object_list = [DEF_SS]
        if subobject_list is None:
            subobject_list = [DEF_ES]
        if builtin_list is None:
            builtin_list = [1, 2, 3, 4, 5]
        if union_arg is None:
            union_arg = [DEF_EES, DEF_ES]
        if array_arg is None:
            array_arg = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]

        StandaloneObject.__init__(
            self, standalone_subobject=standalone_subobject,
            embedded_subobject=embedded_subobject, dynamic_dict=dynamic_dict,
            float_dict=float_dict, string_dict=string_dict,
            tuple_arg=tuple_arg, intarg=intarg, strarg=strarg,
            object_list=object_list, subobject_list=subobject_list,
            builtin_list=builtin_list, union_arg=union_arg,
            subclass_arg=subclass_arg, array_arg=array_arg, name=name
        )


DEF_SOWDV = StandaloneObjectWithDefaultValues()


class Generator(DessiaObject):
    _standalone_in_db = True

    def __init__(self, parameter: int, nb_solutions: int = 25, name: str = ''):
        self.parameter = parameter
        self.nb_solutions = nb_solutions
        self.models = None

        DessiaObject.__init__(self, name=name)

    def generate(self) -> List[StandaloneObject]:
        self.models = [StandaloneObject.generate(self.parameter + i)
                       for i in range(self.nb_solutions)]
        return self.models


class Optimizer(DessiaObject):
    _standalone_in_db = True

    def __init__(self, model_to_optimize: StandaloneObject, name: str = ''):
        self.model_to_optimize = model_to_optimize

        DessiaObject.__init__(self, name=name)

    def optimize(self, optimization_value: int = 3) -> int:
        self.model_to_optimize.intarg += optimization_value
        return self.model_to_optimize.intarg


class Container(DessiaObject):
    _standalone_in_db = True
    _allowed_methods = ["generate_from_text_files"]

    def __init__(self, models: List[StandaloneObject] = None, name: str = ""):
        if models is None:
            self.models = []
        else:
            self.models = models

        DessiaObject.__init__(self, name=name)

    @classmethod
    def generate_from_text_files(cls, files: List[TextIO],
                                 name: str = "Generated from text files"):
        models = [StandaloneObject.generate_from_text(file) for file in files]
        return cls(models=models, name=name)



import dessia_common as dc
from dessia_common.utils.types import is_sequence, is_typing


def deepcopy_value(value, memo):

    if isinstance(value, type) or is_typing(value):  # For class
        return value

    elif isinstance(value, (float, int, str)):
        copied_value = value
        return copied_value

    elif value is None:
        return None

    elif value.__class__.__name__ in ['Point2D', 'Point3D',
                                      'Vector2D', 'Vector3D']:
        copied_value = value.copy(deep=True, memo=memo)
        return copied_value

    elif isinstance(value, dc.DessiaObject):
        memo_value = search_memo(value, memo)
        if memo_value is not None:
            return memo_value
        copied_value = value.copy(deep=True, memo=memo)
        memo[value] = copied_value
        return copied_value

    elif hasattr(value, '__deepcopy__'):
        memo_value = search_memo(value, memo)
        if memo_value is not None:
            return memo_value

        try:
            copied_value = value.__deepcopy__(memo)
        except TypeError:
            copied_value = value.__deepcopy__()
        memo[value] = copied_value
        return copied_value

    else:
        if is_sequence(value):
            return deepcopy_sequence(value, memo)

        elif isinstance(value, dict):
            return deepcopy_dict(value, memo)

        else:
            raise NotImplementedError(
                'unhandle type for copy: {} of type {}'.format(
                    value, value.__class__))


def deepcopy_dict(dict_value, memo):
    memo_value = search_memo(dict_value, memo)
    if memo_value is not None:
        return memo_value

    copied_dict = {}
    for k, v in dict_value.items():
        copied_k = deepcopy_value(k, memo=memo)
        copied_v = deepcopy_value(v, memo=memo)
        copied_dict[copied_k] = copied_v
    return copied_dict


def deepcopy_sequence(seq_value, memo):
    memo_value = search_memo(seq_value, memo)
    if memo_value is not None:
        return memo_value

    copied_list = []
    for v in seq_value:
        cv = deepcopy_value(v, memo=memo)
        copied_list.append(cv)
    return copied_list


def search_memo(value, memo):
    for key in memo.keys():
        if isinstance(value, type(key)) and value == key:
            return memo[value]
    return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""


class ExceptionWithTraceback(Exception):
    def __init__(self, message, traceback_=''):
        self.message = message
        self.traceback = traceback_

    def __str__(self):
        return '{}\nTraceback:\n{}'.format(self.message, self.traceback)


class DeepAttributeError(ExceptionWithTraceback, AttributeError):
    pass


class ModelError(Exception):
    pass


class ConsistencyError(Exception):
    pass


class SerializationError(Exception):
    pass


class DeserializationError(Exception):
    pass


class UntypedArgumentError(Exception):
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov 16 12:25:54 2021

@author: steven
"""

import warnings
import inspect

import dessia_common as dc
import dessia_common.errors as dc_err
import dessia_common.utils.types as dcty
from dessia_common.graph import explore_tree_from_leaves#, cut_tree_final_branches
from dessia_common.breakdown import get_in_object_from_path
import networkx as nx

def serialize_with_pointers(deserialized_element, memo=None, path='#'):
    if memo is None:
        memo = {}
    if isinstance(deserialized_element, dc.DessiaObject):
        try:
            try:
                serialized = deserialized_element.to_dict(use_pointers=True, memo=memo, path=path)
            except TypeError:
                serialized = deserialized_element.to_dict()
                
        except TypeError:
            # warnings.warn('specific to_dict should implement memo and path arguments', Warning)
            serialized, memo = serialize_dict_with_pointers(deserialized_element.to_dict(), memo, path)
            
    elif isinstance(deserialized_element, dict):
        serialized, memo = serialize_dict_with_pointers(deserialized_element, memo, path)
    elif dcty.is_sequence(deserialized_element):
        serialized, memo = serialize_sequence_with_pointers(deserialized_element, memo, path)
    else:
        serialized = deserialized_element
    return serialized, memo


def serialize_dict_with_pointers(dict_, memo, path):
        
    serialized_dict = {}
    dict_attrs_keys = []
    seq_attrs_keys = []
    for key, value in dict_.items():
        value_path = '{}/{}'.format(path, key)
        if hasattr(value, 'to_dict'):
            # object
            if value in memo:
                serialized_dict[key] = {"$ref": memo[value]}
            else:
                try:
                    serialized_dict[key] = value.to_dict(use_pointers=True, path=value_path, memo=memo)
                except TypeError:
                    # warnings.warn('specific to_dict should implement memo and path arguments', Warning)
                    serialized_dict[key] = value.to_dict()
                memo[value] = value_path
        elif isinstance(value, dict):
            dict_attrs_keys.append(key)
        elif isinstance(value, (list, tuple)):
            seq_attrs_keys.append(key)
        else:
            if not dcty.is_jsonable(value):
                msg = 'Attribute {} of value {} is not json serializable'
                raise dc_err.SerializationError(msg.format(key, value))
            serialized_dict[key] = value
        
    # Handle seq & dicts afterwards
    for key in seq_attrs_keys:
        value_path = '{}/{}'.format(path, key)
        serialized_dict[key], memo = serialize_sequence_with_pointers(dict_[key], memo=memo, path=value_path)

    for key in dict_attrs_keys:
        value_path = '{}/{}'.format(path, key)
        serialized_dict[key], memo = serialize_dict_with_pointers(dict_[key], memo=memo, path=value_path)
    return serialized_dict, memo


def serialize_sequence_with_pointers(seq, memo, path):
    serialized_sequence = []
    for ival, value in enumerate(seq):
        value_path = '{}/{}'.format(path, ival)
        if hasattr(value, 'to_dict'):
            if value in memo:
                serialized_value = {"$ref": memo[value]}
            else:
                try:
                    serialized_value = value.to_dict(use_pointers=True, path=value_path, memo=memo)
                except TypeError:
                    # warnings.warn('specific to_dict should implement memo and path arguments', Warning)
                    serialized_value = value.to_dict()
                memo[value] = value_path
            serialized_sequence.append(serialized_value)
        elif isinstance(value, dict):
            serialized_value, memo = serialize_dict_with_pointers(value, memo=memo, path=value_path)
            serialized_sequence.append(serialized_value)
        elif isinstance(value, (list, tuple)):
            serialized_value, memo = serialize_sequence_with_pointers(value, memo=memo, path=value_path)
            serialized_sequence.append(serialized_value)
        else:
            serialized_sequence.append(value)
    return serialized_sequence, memo



def deserialize(serialized_element, sequence_annotation: str = 'List',
                global_dict=None, pointers_memo=None, path='#'):#, enforce_pointers=False):
    # if pointers_memo is None:
    #     pointers_memo = {}
    if pointers_memo is not None:
        if path in pointers_memo:
            return pointers_memo[path]
    
    if isinstance(serialized_element, dict):
        try:
            return dict_to_object(serialized_element, global_dict=global_dict,
                                  pointers_memo=pointers_memo,
                                  path=path)
        except TypeError:
            # warnings.warn('specific dict_to_object of class {}'
            #               ' should implement global_dict and'
            #               ' pointers_memo arguments'.format(serialized_element.__class__.__name__),
            #               Warning)
            return dict_to_object(serialized_element)
    elif dcty.is_sequence(serialized_element):
        return deserialize_sequence(sequence=serialized_element,
                                    annotation=sequence_annotation,
                                    global_dict=global_dict,
                                    pointers_memo=pointers_memo,
                                    path=path)
    return serialized_element

def deserialize_sequence(sequence, annotation=None,
                         global_dict=None, pointers_memo=None,
                         path='#'):
    # TODO: rename to deserialize sequence? Or is this a duplicate ?
    origin, args = dcty.unfold_deep_annotation(typing_=annotation)
    deserialized_sequence = []
    for ie, elt in enumerate(sequence):
        path_elt = '{}/{}'.format(path, ie)
        deserialized_element = deserialize(elt, args,
                                           global_dict=global_dict,
                                           pointers_memo=pointers_memo,
                                           path=path_elt)
        deserialized_sequence.append(deserialized_element)
    if origin is tuple:
        # Keeping as a tuple
        return tuple(deserialized_sequence)
    return deserialized_sequence

def dict_to_object(dict_, class_=None, force_generic: bool = False,
                   global_dict=None, pointers_memo=None, path='#'):
        
    
    if '$ref' in dict_:
        # and dict_['$ref'] in pointers_memo:
        # print(dict_['$ref'])
        # print('This is a ref', path, pointers_memo[dict_['$ref']])
        return pointers_memo[dict_['$ref']]
    
    class_argspec = None

    if pointers_memo is None:
        pointers_memo = {}

    if global_dict is None:
        global_dict = dict_
        pointers_memo.update(dereference_jsonpointers(dict_))
        
        
    working_dict = dict_

    if class_ is None and 'object_class' in working_dict:
        class_ = dcty.get_python_class_from_class_name(working_dict['object_class'])


    if class_ is not None and hasattr(class_, 'dict_to_object'):
        different_methods = (class_.dict_to_object.__func__
                             is not dc.DessiaObject.dict_to_object.__func__)

        if different_methods and not force_generic:
            try:
                obj = class_.dict_to_object(dict_,
                                            global_dict=global_dict,
                                            pointers_memo=pointers_memo)
            except TypeError:
                # warn_msg = 'specific dict_to_object of class {} should implement global_dict arguments'.format(class_.__name__)
                # warnings.warn(warn_msg, Warning)
                obj = class_.dict_to_object(dict_)
            return obj

        if class_._init_variables is None:
            class_argspec = inspect.getfullargspec(class_)
            init_dict = {k: v for k, v in working_dict.items()
                         if k in class_argspec.args}
        else:
            init_dict = {k: v for k, v in working_dict.items()
                         if k in class_._init_variables}
        # TOCHECK Class method to generate init_dict ??
    else:
        init_dict = working_dict
        
    
    subobjects = {}
    for key, value in init_dict.items():
        if class_argspec is not None and key in class_argspec.annotations:
            annotation = class_argspec.annotations[key]
        else:
            annotation = None
        
        key_path = '{}/{}'.format(path, key)
        if key_path in pointers_memo:
            subobjects[key] = pointers_memo[key_path]
        else:
            subobjects[key] = deserialize(value, annotation,
                                          global_dict=global_dict,
                                          pointers_memo=pointers_memo,
                                          path=key_path)#, enforce_pointers=False)

    if class_ is not None:
        obj = class_(**subobjects)
    else:
        obj = subobjects
    
    return obj

def find_references(value, path='#'):
    if isinstance(value, dict):
        return find_references_dict(value, path)
    if dcty.isinstance_base_types(value):
        return []
    elif dcty.is_sequence(value):
        return find_references_sequence(value, path)
    else:
        raise ValueError(value)


def find_references_sequence(seq, path='#'):
    if isinstance(seq, str):
        raise ValueError


    references = []
    for ie, element in enumerate(seq):
        path_value = '{}/{}'.format(path, ie)
        references.extend(find_references(element, path=path_value))
        # if value_nodes or value_edges:
        
    return references

def find_references_dict(dict_, path='#'):
    
    if '$ref' in dict_:
        return [(path, dict_['$ref'])]
    
    references = []
    for key, value in dict_.items():
        if not dcty.isinstance_base_types(value):
            path_value = '{}/{}'.format(path, key)
            references.extend(find_references(value, path=path_value))
    return references


def pointer_graph(value):

    nodes = set()    
    edges = set()
    # print(find_references(value))
    for path, reference in find_references(value):
        segments = path.split('/')
        nodes.add(segments[0])
        previous_node = segments[0]
        for s in segments[1:]:
            node = '{}/{}'.format(previous_node, s)
            nodes.add(node)            
            edges.add((previous_node, node))
            previous_node = node
        edges.add((path, reference))
        # print(path,'->', reference)
            
    
            
    graph = nx.DiGraph()
    graph.add_nodes_from(nodes)
    graph.add_edges_from(edges)
        

    return graph
    

def dereference_jsonpointers(value):#, global_dict):
    graph = pointer_graph(value)
    
    pointers_memo = {}
    if '#' in graph.nodes:
        cycles = list(nx.simple_cycles(graph))
        if cycles:
            for cycle in cycles:
                print(cycle)
            raise NotImplementedError('Cycles in ref not handled')
            
        order = list(explore_tree_from_leaves(graph))
        if '#' in order:
            order.remove('#')
            
        # for ref in order:
        #     print('ref', ref)

        for ref in order:
            serialized_element = get_in_object_from_path(value, ref)
            # print(serialized_element)
            # try:
            pointers_memo[ref] = deserialize(serialized_element=serialized_element,
                                     global_dict=value,
                                     pointers_memo=pointers_memo,
                                     path=ref)
            # except:
            #     print(ref)
            #     # print('\n', serialized_element)
            #     # print('\n\n', pointers_memo)
            #     return serialized_element, pointers_memo
            #     raise RuntimeError('jjj')
            # print('\nref', ref, pointers_memo[ref])
    # print(pointers_memo.keys())
    return pointers_memo
            




def pointer_graph_elements(value, path='#'):
    # edges = []
    # nodes = []

    if isinstance(value, dict):
        return pointer_graph_elements_dict(value, path)
    if dcty.isinstance_base_types(value):
        return [], []
    elif dcty.is_sequence(value):
        return pointer_graph_elements_sequence(value, path)
    else:
        raise ValueError(value)



def pointer_graph_elements_sequence(seq, path='#'):
    if isinstance(seq, str):
        raise ValueError


    edges = []
    nodes = []
    for ie, element in enumerate(seq):
        path_value = '{}/{}'.format(path, ie)
        value_nodes, value_edges = pointer_graph_elements(element, path=path_value)
        # if value_nodes or value_edges:
        
        nodes.append(path_value)
        nodes.extend(value_nodes)
        
        edges.append((path, path_value))
        edges.extend(value_edges)


    return nodes, edges

def pointer_graph_elements_dict(dict_, path='#'):
    
    
    if '$ref' in dict_:
        return [path, dict_['$ref']], [(path, dict_['$ref'])]
    
    edges = []
    nodes = []
    for key, value in dict_.items():
        if not dcty.isinstance_base_types(value):
            path_value = '{}/{}'.format(path, key)
            value_nodes, value_edges = pointer_graph_elements(value, path=path_value)
            # if value_nodes or value_edges:        
            nodes.append(path_value)
            nodes.extend(value_nodes)
        
            edges.append((path, path_value))
            edges.extend(value_edges)

    return nodes, edges


def pointers_analysis(obj):
    if isinstance(obj, dict):
        dict_ = obj
    else:
        dict_ = obj.to_dict()
    
    class_number = {}
    composed_by = {}
    class_from_path = {}
    graph = pointer_graph(dict_)
    for path1, path2 in graph.edges():
        if path1 != '#':
            # print(path1, path2)
            if path2 in class_from_path:
                val2_class = class_from_path[path2]
            else:
                val2 = get_in_object_from_path(obj, path2)
                val2_class = dcty.full_classname(val2)
                # val2_class = val2['object_class']
                class_from_path[path2] = val2_class
    
            # if 'object_class' in val2:
            # val2_class = val2['object_class']
            class_number[val2_class] = class_number.get(val2_class, 0) + 1
    
            
            if path1 in class_from_path:
                val1_class = class_from_path[path1]
            else:
                val1 = get_in_object_from_path(obj, path1)
                val1_class = dcty.full_classname(val1)
                # val1_class = val1['object_class']
                class_from_path[path1] = val1_class
    
            
            if val1_class != val2_class:
                if not val2_class in composed_by:
                    composed_by[val2_class] = {}
                
                if not val1_class in composed_by[val2_class]:
                    composed_by[val2_class][val1_class] = 1
                else:
                    composed_by[val2_class][val1_class] += 1
                
    return class_number, composed_by
                
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""
import os
import inspect
import time
import tempfile
import json
import webbrowser
import networkx as nx
from typing import List, Union, Type, Any, Dict,\
    Tuple, Optional, get_type_hints
from copy import deepcopy
from dessia_common.templates import workflow_template
import itertools

from dessia_common import DessiaObject, DisplayObject, DessiaFilter, \
    is_sequence, list_hash, serialize, is_bounded, \
    type_from_annotation,\
    enhanced_deep_attr, deprecation_warning, JSONSCHEMA_HEADER,\
    jsonschema_from_annotation, deserialize_argument, set_default_value,\
    prettyname, serialize_dict,\
    recursive_type, recursive_instantiation
from dessia_common.errors import UntypedArgumentError
from dessia_common.utils.diff import data_eq
from dessia_common.utils.serialization import dict_to_object, deserialize,\
    serialize_with_pointers
from dessia_common.utils.types import get_python_class_from_class_name,\
    serialize_typing, full_classname, deserialize_typing
from dessia_common.utils.copy import deepcopy_value
from dessia_common.vectored_objects import from_csv
from dessia_common.typings import JsonSerializable, MethodType,\
    ClassMethodType
import warnings

# Type Aliases
VariableTypes = Union['Variable', 'TypedVariable',
                      'VariableWithDefaultValue',
                      'TypedVariableWithDefaultValue']


# VariableTypes = Subclass['Variable']


class Variable(DessiaObject):
    _standalone_in_db = False
    _eq_is_data_eq = False
    has_default_value: bool = False

    def __init__(self, memorize: bool = False, name: str = ''):
        self.memorize = memorize
        DessiaObject.__init__(self, name=name)
        self.position = None

    def to_dict(self):
        dict_ = DessiaObject.base_dict(self)
        dict_.update({'has_default_value': self.has_default_value})
        return dict_


class TypedVariable(Variable):
    has_default_value: bool = False

    def __init__(self, type_: Type, memorize: bool = False, name: str = ''):
        Variable.__init__(self, memorize=memorize, name=name)
        self.type_ = type_

    def to_dict(self):
        dict_ = DessiaObject.base_dict(self)
        dict_.update({'type_': serialize_typing(self.type_),
                      'memorize': self.memorize,
                      'has_default_value': self.has_default_value})
        return dict_

    @classmethod
    def dict_to_object(cls, dict_):
        type_ = deserialize_typing(dict_['type_'])
        memorize = dict_['memorize']
        return cls(type_=type_, memorize=memorize, name=dict_['name'])

    def copy(self, deep: bool = False, memo=None):
        return TypedVariable(type_=self.type_, memorize=self.memorize,
                             name=self.name)


class VariableWithDefaultValue(Variable):
    has_default_value: bool = True

    def __init__(self, default_value: Any, memorize: bool = False,
                 name: str = ''):
        Variable.__init__(self, memorize=memorize, name=name)
        self.default_value = default_value


class TypedVariableWithDefaultValue(TypedVariable):
    has_default_value: bool = True

    def __init__(self, type_: Type, default_value: Any,
                 memorize: bool = False, name: str = ''):
        TypedVariable.__init__(self, type_=type_, memorize=memorize, name=name)
        self.default_value = default_value

    def to_dict(self):
        dict_ = DessiaObject.base_dict(self)
        dict_.update({'type_': serialize_typing(self.type_),
                      'default_value': serialize(self.default_value),
                      'memorize': self.memorize,
                      'has_default_value': self.has_default_value})
        return dict_

    @classmethod
    def dict_to_object(cls, dict_):
        type_ = get_python_class_from_class_name(dict_['type_'])
        default_value = deserialize(dict_['default_value'])
        return cls(type_=type_, default_value=default_value,
                   memorize=dict_['memorize'], name=dict_['name'])

    def copy(self, deep: bool = False, memo=None):
        copied_default_value = deepcopy_value(self.default_value, memo=memo)
        return TypedVariableWithDefaultValue(type_=self.type_,
                                             default_value=copied_default_value,
                                             memorize=self.memorize,
                                             name=self.name)


def set_block_variable_names_from_dict(func):
    def func_wrapper(cls, dict_):
        obj = func(cls, dict_)
        if 'input_names' in dict_:
            for input_name, input_ in zip(dict_['input_names'], obj.inputs):
                input_.name = input_name
        if 'output_names' in dict_:
            output_items = zip(dict_['output_names'], obj.outputs)
            for output_name, output_ in output_items:
                output_.name = output_name
        return obj

    return func_wrapper


class Block(DessiaObject):
    _standalone_in_db = False
    _eq_is_data_eq = False
    _non_serializable_attributes = []

    def __init__(self, inputs: List[VariableTypes],
                 outputs: List[VariableTypes],
                 position: Tuple[float, float] = None, name: str = ''):
        self.inputs = inputs
        self.outputs = outputs
        if position is None:
            self.position = (0, 0)
        else:
            self.position = position

        DessiaObject.__init__(self, name=name)

    def equivalent_hash(self):
        return len(self.__class__.__name__)

    def equivalent(self, other):
        return self.__class__.__name__ == other.__class__.__name__

    def to_dict(self):
        dict_ = DessiaObject.base_dict(self)
        dict_['inputs'] = [i.to_dict() for i in self.inputs]
        dict_['outputs'] = [o.to_dict() for o in self.outputs]
        if self.position is not None:
            dict_['position'] = list(self.position)
        else:
            dict_['position'] = self.position
        return dict_

    def jointjs_data(self):
        data = {'block_class': self.__class__.__name__}
        if self.name != '':
            data['name'] = self.name
        else:
            data['name'] = self.__class__.__name__
        return data


class Display(Block):
    _displayable_input = 0
    _non_editable_attributes = ['inputs']

    def __init__(self, inputs: List[VariableTypes] = None, order: int = 0, name: str = ''):
        self.order = order
        if inputs is None:
            inputs = [Variable(name='Model to Display', memorize=True)]

        Block.__init__(self, inputs=inputs, outputs=[], name=name)

    def equivalent(self, other):
        if not Block.equivalent(self, other):
            return False
        return self.order == other.order

    def equivalent_hash(self):
        return self.order

    def display_(self, local_values: Dict[VariableTypes, Any], **kwargs):
        object_ = local_values[self.inputs[self._displayable_input]]
        displays = object_._displays(**kwargs)
        return displays

    def to_dict(self):
        dict_ = Block.to_dict(self)
        dict_['order'] = self.order
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_):
        return cls(order=dict_['order'], name=dict_['name'])

    @staticmethod
    def evaluate(self):
        return []


class Import(Block):
    def __init__(self, type_: str, name: str = ''):
        self.type_ = type_
        inputs = [TypedVariable(type_=str, name='Input filename'),
                  TypedVariableWithDefaultValue(type_=bool, default_value=True,
                                                name='Remove duplicates')]
        outputs = [Variable(name='Array'), Variable(name='Variables')]

        Block.__init__(self, inputs=inputs, outputs=outputs, name=name)

    def equivalent_hash(self):
        return len(self.type_)

    def equivalent(self, other):
        if not Block.equivalent(self, other):
            return False
        return self.type_ == other.type_

    def to_dict(self):
        dict_ = Block.to_dict(self)
        dict_['type_'] = self.type_
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_):
        return cls(type_=dict_['type_'], name=dict_['name'])

    def evaluate(self, values):
        dirname = os.path.dirname(__file__)
        relative_filepath = 'models/data/' + values[self.inputs[0]]
        filename = os.path.join(dirname, relative_filepath)
        if self.type_ == 'csv':
            array, variables = from_csv(filename=filename, end=None,
                                        remove_duplicates=True)
            return [array, variables]
        msg = 'File type {} not supported'.format(self.type_)
        raise NotImplementedError(msg)


class InstantiateModel(Block):
    """
    :param model_class: The class to instanciate.
    :type model_class: Instanciable
    :param name: The name of the block.
    :type name: str
    """

    def __init__(self, model_class: Type, name: str = ''):
        self.model_class = model_class
        inputs = []

        inputs = set_inputs_from_function(self.model_class.__init__, inputs)

        outputs = [TypedVariable(type_=self.model_class,
                                 name='Instanciated object')]
        Block.__init__(self, inputs, outputs, name=name)

    def equivalent_hash(self):
        return len(self.model_class.__name__)

    def equivalent(self, other):
        if not Block.equivalent(self, other):
            return False
        classname = self.model_class.__class__.__name__
        other_classname = other.model_class.__class__.__name__
        return classname == other_classname

    def to_dict(self):
        dict_ = Block.to_dict(self)
        dict_['model_class'] = full_classname(object_=self.model_class,
                                              compute_for='class')
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_):
        if 'model_class_module' in dict_:
            # TODO Retro-compatibility. Remove this in future versions
            module_name = dict_['model_class_module']
            classname = module_name + '.' + dict_['model_class']
        else:
            classname = dict_['model_class']
        class_ = get_python_class_from_class_name(classname)
        return cls(class_, name=dict_['name'])

    def evaluate(self, values):
        args = {var.name: values[var] for var in self.inputs}
        return [self.model_class(**args)]

    def package_mix(self):
        return {self.model_class.__module__.split('.')[0]: 1}


class InstanciateModel(InstantiateModel):
    def __init__(self, model_class: Type, name: str = ''):
        InstantiateModel.__init__(self, model_class=model_class, name=name)
        warnings.warn(
            "InstanciateModel is deprecated, use InstantiateModel instead",
            DeprecationWarning
        )


class ClassMethod(Block):
    def __init__(self, method_type: ClassMethodType[Type], name: str = ''):
        self.method_type = method_type
        inputs = []
        method = getattr(method_type.class_, method_type.name)
        inputs = set_inputs_from_function(method, inputs)

        self.argument_names = [i.name for i in inputs]

        output_name = 'method result of {}'.format(method_type.name)
        annotations = get_type_hints(method)
        if 'return' in annotations:
            type_ = type_from_annotation(annotations['return'],
                                         method.__module__)
            outputs = [TypedVariable(type_=type_, name=output_name)]
        else:
            outputs = [Variable(name=output_name)]
        Block.__init__(self, inputs, outputs, name=name)

    def equivalent_hash(self):
        classname = self.method_type.class_.__name__
        return len(classname) + 7 * len(self.method_type.name)

    def equivalent(self, other: 'ClassMethod'):
        if not Block.equivalent(self, other):
            return False
        classname = self.method_type.class_.__name__
        other_classname = other.method_type.class_.__name__
        same_class = classname == other_classname
        same_method = self.method_type.name == other.method_type.name
        return same_class and same_method

    def to_dict(self):
        dict_ = Block.to_dict(self)
        classname = full_classname(object_=self.method_type.class_,
                                   compute_for='class')
        method_type_dict = {'class_': classname, 'name': self.method_type.name}
        dict_.update({'method_type': method_type_dict})
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_) -> 'ClassMethod':
        if 'method_type' in dict_:
            classname = dict_['method_type']['class_']
            method_name = dict_['method_type']['name']
        else:
            # Retro-compatibility
            classname = dict_['model_class']
            method_name = dict_['method_name']
        class_ = get_python_class_from_class_name(classname)
        name = dict_['name']
        method_type = ClassMethodType(class_=class_, name=method_name)
        return cls(method_type=method_type, name=name)

    def evaluate(self, values):
        args = {arg_name: values[var]
                for arg_name, var in zip(self.argument_names, self.inputs)
                if var in values}
        return [getattr(self.method_type.class_,
                        self.method_type.name)(**args)]


class ModelMethod(Block):
    """
    :param method_type: Represent class and method used.
    :type method_type: MethodType[T]
    :param name: The name of the block.
    :type name: str
    """

    def __init__(self, method_type: MethodType[Type], name: str = ''):
        self.method_type = method_type
        inputs = [TypedVariable(type_=method_type.class_,
                                name='model at input')]
        method = getattr(method_type.class_, method_type.name)

        inputs = set_inputs_from_function(method, inputs)

        # Storing argument names
        self.argument_names = [i.name for i in inputs[1:]]

        result_output_name = 'method result of {}'.format(method_type.name)
        annotations = get_type_hints(method)
        if 'return' in annotations:
            type_ = type_from_annotation(annotations['return'],
                                         method.__module__)
            return_output = TypedVariable(type_=type_, name=result_output_name)
        else:
            return_output = Variable(name=result_output_name)

        model_output_name = 'model at output {}'.format(method_type.name)
        model_output = TypedVariable(type_=method_type.class_,
                                     name=model_output_name)
        outputs = [return_output, model_output]
        if name == '':
            name = 'Model method: {}'.format(method_type.name)
        Block.__init__(self, inputs, outputs, name=name)

    def equivalent_hash(self):
        classname = self.method_type.class_.__name__
        return len(classname) + 7 * len(self.method_type.name)

    def equivalent(self, other: 'ModelMethod'):
        if not Block.equivalent(self, other):
            return False
        classname = self.method_type.class_.__name__
        other_classname = other.method_type.class_.__name__
        same_model = classname == other_classname
        same_method = self.method_type.name == other.method_type.name
        return same_model and same_method

    def to_dict(self):
        dict_ = Block.to_dict(self)
        classname = full_classname(object_=self.method_type.class_,
                                   compute_for='class')
        method_type_dict = {'class_': classname, 'name': self.method_type.name}
        dict_.update({'method_type': method_type_dict})
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_) -> 'ModelMethod':
        if 'method_type' in dict_:
            classname = dict_['method_type']['class_']
            method_name = dict_['method_type']['name']
        else:
            # Retro-compatibility
            classname = dict_['model_class']
            method_name = dict_['method_name']
        class_ = get_python_class_from_class_name(classname)
        name = dict_['name']
        method_type = MethodType(class_=class_, name=method_name)
        return cls(method_type=method_type, name=name)

    def evaluate(self, values):
        args = {arg_name: values[var]
                for arg_name, var in zip(self.argument_names, self.inputs[1:])
                if var in values}
        return [getattr(values[self.inputs[0]], self.method_type.name)(**args),
                values[self.inputs[0]]]

    def package_mix(self):
        return {self.method_type.class_.__module__.split('.')[0]: 1}


# class Function(Block):
#     def __init__(self, function: Callable, name: str = ''):
#         self.function = function
#         inputs = []
#         annotations = get_type_hints(function)
#         for arg_name in inspect.signature(function).parameters.keys():
#             # TODO: Check why we need TypedVariables
#             type_ = type_from_annotation(annotations[arg_name])
#             inputs.append(TypedVariable(type_=type_, name=arg_name))
#         out_type = type_from_annotation(annotations['return'])
#         outputs = [TypedVariable(type_=out_type, name='Output function')]
#
#         Block.__init__(self, inputs, outputs, name=name)
#
#     def equivalent_hash(self):
#         return int(hash(self.function.__name__) % 10e5)
#
#     def equivalent(self, other):
#         return self.function == other.function
#
#     def evaluate(self, values):
#         return self.function(*values)


class Sequence(Block):
    def __init__(self, number_arguments: int, name: str = ''):
        # type_: Subclass[DessiaObject] = None,
        self.number_arguments = number_arguments
        prefix = 'Sequence element {}'
        inputs = [Variable(name=prefix.format(i))
                  for i in range(self.number_arguments)]
        # if type_ is None:
        # else:
        #     inputs = [TypedVariable(type_=type_, name=prefix.format(i))
        #               for i in range(self.number_arguments)]

        # self.type_ = type_
        outputs = [TypedVariable(type_=list, name='sequence')]
        Block.__init__(self, inputs, outputs, name=name)

    def equivalent_hash(self):
        return self.number_arguments

    def equivalent(self, other):
        if not Block.equivalent(self, other):
            return False
        return self.number_arguments == other.number_arguments

    def to_dict(self):
        dict_ = Block.to_dict(self)
        dict_['number_arguments'] = self.number_arguments
        # if self.type_ is not None:
        #     dict_['type_'] = serialize_typing(self.type_)
        # else:
        #     dict_['type_'] = None
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_):
        # if dict_['type_'] is not None:
        #     type_ = deserialize_typing(dict_['type_'])
        # else:
        #     type_ = None
        return cls(dict_['number_arguments'], dict_['name'])

    def evaluate(self, values):
        return [[values[var] for var in self.inputs]]


class ForEach(Block):
    """
    :param workflow_block: The WorkflowBlock on which iterate.
    :type workflow_block: WorkflowBlock
    :param iter_input_index: Index of iterable input in worklow_block.inputs
    :type iter_input_index: int
    :param input_connections: Links ForEach's inputs to its
        workflow_block's inputs.
        input_connections[i] = [ForEach_input_j, WorkflowBlock_input_k]
    :type input_connections: List[Tuple[Tuple[int, int, int], Tuple[int, int, int]]]
    :param output_connections: Same but for outputs.
        output_connections[i] = [WorkflowBlock_output_j, ForEach_output_k]
    :type output_connections: List[Tuple[Tuple[int, int, int], Tuple[int, int, int]]]
    :param name: The name of the block.
    :type name: str
    """

    def __init__(self, workflow_block: 'WorkflowBlock',
                 iter_input_index: int, name: str = ''):
        self.workflow_block = workflow_block
        self.iter_input_index = iter_input_index
        self.iter_input = self.workflow_block.inputs[iter_input_index]
        inputs = []
        for i, workflow_input in enumerate(self.workflow_block.inputs):
            if i == iter_input_index:
                name = 'Iterable input: ' + workflow_input.name
                inputs.append(Variable(name=name))
            else:
                input_ = workflow_input.copy()
                input_.name = 'binding ' + input_.name
                inputs.append(input_)
        output_variable = Variable(name='Foreach output')
        self.output_connections = None  # TODO: configuring port internal connections
        self.input_connections = None

        Block.__init__(self, inputs, [output_variable], name=name)

    def equivalent_hash(self):
        wb_hash = int(self.workflow_block.equivalent_hash() % 10e5)
        return wb_hash + self.iter_input_index

    def equivalent(self, other):
        if not Block.equivalent(self, other):
            return False
        input_eq = self.iter_input_index == other.iter_input_index
        wb_eq = self.workflow_block.equivalent(other.workflow_block)
        return wb_eq and input_eq

    def to_dict(self):
        dict_ = Block.to_dict(self)
        dict_.update({'workflow_block': self.workflow_block.to_dict(),
                      'iter_input_index': self.iter_input_index})
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_):
        workflow_block = WorkflowBlock.dict_to_object(dict_['workflow_block'])
        iter_input_index = dict_['iter_input_index']
        return cls(workflow_block=workflow_block,
                   iter_input_index=iter_input_index, name=dict_['name'])

    def evaluate(self, values):
        values_workflow = {var2: values[var1]
                           for var1, var2 in zip(self.inputs,
                                                 self.workflow_block.inputs)}
        output_values = []
        for value in values_workflow[self.iter_input]:
            values_workflow[self.iter_input] = value
            output = self.workflow_block.evaluate(values_workflow)[0]
            output_values.append(output)
        return [output_values]


class Unpacker(Block):
    def __init__(self, indices: List[int], name: str = ''):
        self.indices = indices
        inputs = [Variable(name='input_sequence')]
        outputs = [Variable(name='output_{}'.format(i)) for i in indices]

        Block.__init__(self, inputs=inputs, outputs=outputs, name=name)

    def equivalent(self, other):
        if not Block.equivalent(self, other):
            return False
        return self.indices == other.indices

    def equivalent_hash(self):
        return len(self.indices)

    def to_dict(self):
        dict_ = Block.to_dict(self)
        dict_['indices'] = self.indices
        return dict_

    @classmethod
    def dict_to_object(cls, dict_):
        return cls(dict_['indices'], dict_['name'])

    def evaluate(self, values):
        return [values[self.inputs[0]][i] for i in self.indices]


class Flatten(Block):
    def __init__(self, name: str = ''):
        inputs = [Variable(name='input_sequence')]
        outputs = [Variable(name='flatten_sequence')]
        Block.__init__(self, inputs, outputs, name=name)

    def equivalent_hash(self):
        return 1

    @classmethod
    def dict_to_object(cls, dict_):
        return cls(dict_['name'])

    def evaluate(self, values):
        output = []
        for value in values[self.inputs[0]]:
            output.extend(value)
        return [output]


class Product(Block):
    def __init__(self, number_list: int, name=''):
        self.number_list = number_list
        inputs = []
        for i in range(self.number_list):
            inputs.append(Variable(name='list_product_' + str(i)))

        output_variable = Variable(name='Product output')
        Block.__init__(self, inputs, [output_variable], name=name)

    def equivalent_hash(self):
        return self.number_list

    def equivalent(self, other):
        if not Block.equivalent(self, other):
            return False
        return self.number_list == other.number_list

    def to_dict(self):
        dict_ = DessiaObject.base_dict(self)
        dict_.update({'number_list': self.number_list})
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_):
        number_list = dict_['number_list']
        return cls(number_list=number_list, name=dict_['name'])

    def evaluate(self, values):
        list_product = [values[var] for var in self.inputs]
        output_value = list(itertools.product(*list_product))
        return [output_value]


class Filter(Block):
    """
    :param filters: A list of dictionaries,
                    each corresponding to a value to filter.
                    The dictionary should be as follows :
                    *{'attribute' : Name of attribute to filter (str),
                      'operator' : choose between gt, lt, get, let
                                  (standing for greater than, lower than,
                                   geater or equal than,
                                   lower or equal than) (str),
                      'bound' :  the value (float)}*
    :type filters: list[DessiaFilter]
    :param name: The name of the block.
    :type name: str
    """

    def __init__(self, filters: List[DessiaFilter], name: str = ''):
        self.filters = filters
        inputs = [Variable(name='input_list')]
        outputs = [Variable(name='output_list')]
        Block.__init__(self, inputs, outputs, name=name)

    def equivalent(self, other):
        if not Block.equivalent(self, other):
            return False
        return self.filters == other.filters

    def equivalent_hash(self):
        hashes = [hash(f) for f in self.filters]
        return int(sum(hashes) % 10e5)

    def to_dict(self):
        dict_ = Block.to_dict(self)
        dict_.update({'filters': [f.to_dict() for f in self.filters]})
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_):
        return cls([DessiaFilter.dict_to_object(d) for d in dict_['filters']], dict_['name'])

    def evaluate(self, values):
        ouput_values = []
        objects = values[self.inputs[0]]
        for object_ in objects:
            bounded = True
            i = 0
            while bounded and i < len(self.filters):
                filter_ = self.filters[i]
                value = enhanced_deep_attr(object_, filter_.attribute)
                bounded = is_bounded(filter_, value)
                i += 1

            if bounded:
                ouput_values.append(object_)
        return [ouput_values]


class MultiPlot(Display):
    """
    :param attributes: A List of all attributes that will be shown inside the \
    ParallelPlot window on DessIA's Platform.
    :type attributes: List[str]
    :param name: The name of the block.
    :type name: str
    """

    def __init__(self, attributes: List[str], order: int = 0, name: str = ''):
        self.attributes = attributes
        Display.__init__(self, order=order, name=name)
        self.inputs[0].name = 'Input List'

    def equivalent(self, other):
        if not Block.equivalent(self, other):
            return False

        same_attributes = self.attributes == other.attributes
        same_order = self.order == other.order
        return same_attributes and same_order

    def equivalent_hash(self):
        return sum([len(a) for a in self.attributes]) + self.order

    def display_(self, local_values, **kwargs):
        import plot_data
        if 'reference_path' not in kwargs:
            reference_path = 'output_value'  # TODO bof bof bof
        else:
            reference_path = kwargs['reference_path']
        display_input = self.inputs[self._displayable_input]
        objects = local_values[display_input]
        # pareto_settings = local_values[self.inputs[1]]

        values = [{a: enhanced_deep_attr(o, a) for a in self.attributes}
                  for o in objects]

        first_vars = self.attributes[:2]
        values2d = [{key: val[key]} for key in first_vars for val in
                    values]

        tooltip = plot_data.Tooltip(name='Tooltip', attributes=self.attributes)

        scatterplot = plot_data.Scatter(tooltip=tooltip,
                                        x_variable=first_vars[0],
                                        y_variable=first_vars[1],
                                        elements=values2d,
                                        name='Scatter Plot')

        rgbs = [[192, 11, 11], [14, 192, 11], [11, 11, 192]]
        parallelplot = plot_data.ParallelPlot(
            disposition='horizontal', axes=self.attributes,
            rgbs=rgbs, elements=values
        )
        objects = [scatterplot, parallelplot]
        sizes = [plot_data.Window(width=560, height=300),
                 plot_data.Window(width=560, height=300)]
        coords = [(0, 0), (0, 300)]
        multiplot = plot_data.MultiplePlots(elements=values, plots=objects,
                                            sizes=sizes, coords=coords,
                                            name='Results plot')
        display_ = DisplayObject(type_='plot_data', data=multiplot,
                                 reference_path=reference_path)
        return [display_.to_dict()]

    def to_dict(self):
        dict_ = Block.to_dict(self)
        dict_.update({'attributes': self.attributes, 'order': self.order})
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_):
        return cls(attributes=dict_['attributes'],
                   order=dict_['order'],
                   name=dict_['name'])

    @staticmethod
    def evaluate(_):
        return []


class ParallelPlot(MultiPlot):
    """

    DEPRECATED. USE MULTIPLOT INSTEAD

    :param attributes: A List of all attributes that will be shown inside the \
    ParallelPlot window on the DessIA Platform.
    :type attributes: List[str]
    :param name: The name of the block.
    :type name: str
    """

    def __init__(self, attributes: List[str], order: int = 0, name: str = ''):
        deprecation_warning(self.__class__.__name__, 'Class', 'MultiPlot')
        MultiPlot.__init__(self, attributes=attributes, order=order, name=name)

    def to_dict(self):
        dict_ = MultiPlot.to_dict(self)
        dict_.update({'object_class': 'dessia_common.workflow.MultiPlot'})
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_):
        return MultiPlot(attributes=dict_['attributes'], order=dict_['order'],
                         name=dict_['name'])


class ModelAttribute(Block):
    """
    :param attribute_name: The name of the attribute to select.
    :type attribute_name: str
    :param name: The name of the block.
    :type name: str
    """

    def __init__(self, attribute_name: str, name: str = ''):
        self.attribute_name = attribute_name

        inputs = [Variable(name='Model')]
        outputs = [Variable(name='Model attribute')]
        Block.__init__(self, inputs, outputs, name=name)

    def equivalent_hash(self):
        return len(self.attribute_name)

    def equivalent(self, other):
        if not Block.equivalent(self, other):
            return False
        return self.attribute_name == other.attribute_name

    def to_dict(self):
        dict_ = Block.to_dict(self)
        dict_.update({'attribute_name': self.attribute_name})
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_):
        return cls(dict_['attribute_name'], dict_['name'])

    def evaluate(self, values):
        return [enhanced_deep_attr(values[self.inputs[0]],
                                   self.attribute_name)]


class Sum(Block):
    def __init__(self, number_elements: int = 2, name: str = ''):
        self.number_elements = number_elements
        inputs = [Variable(name='Sum element {}'.format(i + 1))
                  for i in range(number_elements)]
        outputs = [Variable(name='Sum')]
        Block.__init__(self, inputs, outputs, name=name)

    def equivalent_hash(self):
        return self.number_elements

    def equivalent(self, other):
        if not Block.equivalent(self, other):
            return False
        return self.number_elements == other.number_elements

    def to_dict(self):
        dict_ = Block.to_dict(self)
        dict_.update({'number_elements': self.number_elements})
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_):
        return cls(dict_['number_elements'], dict_['name'])

    @staticmethod
    def evaluate(values):
        return [sum(values)]


class Substraction(Block):
    def __init__(self, name: str = ''):
        inputs = [Variable(name='+'), Variable(name='-')]
        outputs = [Variable(name='Substraction')]
        Block.__init__(self, inputs, outputs, name=name)

    def equivalent_hash(self):
        return 0

    def equivalent(self, other):
        if not Block.equivalent(self, other):
            return False
        return True

    def to_dict(self):
        dict_ = Block.to_dict(self)
        return dict_

    @classmethod
    def dict_to_object(cls, dict_):
        return cls(dict_['name'])

    def evaluate(self, values):
        return [values[self.inputs[0]] - values[self.inputs[1]]]


class Pipe(DessiaObject):
    """
    :param input_variable: The input varaible of the pipe correspond to the \
    start of the arrow, its tail.
    :type input_variable: TypedVariable
    :param output_variable: The output variable of the pipe correpond to the \
    end of the arrow, its hat.
    :type output_variable: TypedVariable
    """
    _jsonschema = {
        "definitions": {},
        "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "object",
        "title": "Block",
        "python_typing": 'dessia_common.workflow.Pipe',
        "standalone_in_db": False,
        "required": ["input_variable", "output_variable"],
        "properties": {
            "input_variable": {
                "type": "object", "editable": True, "order": 0,
                "python_typing": "List[dessia_common.workflow.VariableTypes]",
                "classes": [
                    "dessia_common.workflow.Variable",
                    "dessia_common.workflow.TypedVariable",
                    "dessia_common.workflow.VariableWithDefaultValue",
                    "dessia_common.workflow.TypedVariableWithDefaultValue"
                ]
            },
            "output_variable": {
                "type": "object", "editable": True, "order": 1,
                "python_typing": "List[dessia_common.workflow.VariableTypes]",
                "classes": [
                    "dessia_common.workflow.Variable",
                    "dessia_common.workflow.TypedVariable",
                    "dessia_common.workflow.VariableWithDefaultValue",
                    "dessia_common.workflow.TypedVariableWithDefaultValue"
                ],
            },
            "name": {
                'type': 'string', 'title': 'Name', 'editable': True,
                'order': 2, 'default_value': '',
                'python_typing': 'builtins.str'
            }
        }
    }
    _eq_is_data_eq = False

    def __init__(self, input_variable: VariableTypes,
                 output_variable: VariableTypes, name: str = ''):
        self.input_variable = input_variable
        self.output_variable = output_variable

        DessiaObject.__init__(self, name=name)

    def to_dict(self):
        return {'input_variable': self.input_variable,
                'output_variable': self.output_variable}


class WorkflowError(Exception):
    pass


class Workflow(Block):
    """
    :param blocks: A List with all the Blocks used by the Worklow.
    :type blocks: List of Block objects
    :param pipes: A List of Pipe objects.
    :type pipes: List of Pipe objects
    :param imposed_variable_values: A dictionary of imposed variable values.
    :type imposed_variable_values: dict
    :param name: The name of the block.
    :type name: str
    """
    _standalone_in_db = True
    _allowed_methods = ['run']
    _eq_is_data_eq = True

    _jsonschema = {
        "definitions": {},
        "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "object",
        "title": "Workflow",
        "required": ["blocks", "pipes", "outputs"],
        "python_typing": 'dessia_common.workflow.Pipe',
        "standalone_in_db": True,
        "properties": {
            "blocks": {
                "type": "array", "order": 0, "editable": True,
                "python_typing": "SubclassOf[dessia_common.workflow.Block]",
                "items": {
                    "type": "object",
                    "classes": ["dessia_common.workflow.InstanciateModel",
                                "dessia_common.workflow.ModelMethod",
                                "dessia_common.workflow.ForEach",
                                "dessia_common.workflow.ModelAttribute",
                                "dessia_common.workflow.Function",
                                "dessia_common.workflow.Sequence",
                                "dessia_common.workflow.ForEach",
                                "dessia_common.workflow.Unpacker",
                                "dessia_common.workflow.Flatten",
                                "dessia_common.workflow.Filter",
                                "dessia_common.workflow.ParallelPlot",
                                "dessia_common.workflow.Sum",
                                "dessia_common.workflow.Substraction"],
                    "editable": True,
                },
            },
            "pipes": {
                "type": "array",
                "order": 1,
                "editable": True,
                "python_typing": "List[dessia_common.workflow.Pipe]",
                "items": {
                    'type': 'objects',
                    'classes': ["dessia_common.workflow.Pipe"],
                    "python_type": "dessia_common.workflow.Pipe",
                    "editable": True
                }
            },
            "outputs": {
                "type": "array", "order": 2,
                "python_typing": "List[dessia_common.workflow.VariableTypes]",
                'items': {
                    'type': 'array',
                    'items': {'type': 'number'},
                    'python_typing': "dessia_common.workflow.VariableTypes"
                }
            },
            "name": {
                'type': 'string', 'title': 'Name', 'editable': True,
                'order': 3, 'default_value': '',
                'python_typing': 'builtins.str'
            }
        }
    }

    def __init__(self, blocks, pipes, output, *,
                 imposed_variable_values=None, name=''):
        self.blocks = blocks
        self.pipes = pipes

        if imposed_variable_values is None:
            self.imposed_variable_values = {}
        else:
            self.imposed_variable_values = imposed_variable_values

        self.coordinates = {}

        self.nonblock_variables = []
        self.variables = []
        for block in self.blocks:
            if isinstance(block, Workflow):
                msg = "Using workflow as blocks is forbidden, "
                msg += "use WorkflowBlock wrapper instead"
                raise ValueError(msg)
            self.variables.extend(block.inputs)
            self.variables.extend(block.outputs)
            try:
                self.coordinates[block] = (0, 0)
            except ValueError:
                msg = "can't serialize block {} ({})".format(block, block.name)
                raise ValueError(msg)

        for pipe in self.pipes:
            upstream_var = pipe.input_variable
            downstream_var = pipe.output_variable
            if upstream_var not in self.variables:
                self.variables.append(upstream_var)
                self.nonblock_variables.append(upstream_var)
            if downstream_var not in self.variables:
                self.variables.append(downstream_var)
                self.nonblock_variables.append(downstream_var)

        self._utd_graph = False

        input_variables = []

        for variable in self.variables:
            if (variable not in self.imposed_variable_values) and \
                    (len(nx.ancestors(self.graph, variable)) == 0):
                # !!! Why not just : if nx.ancestors(self.graph, variable) ?
                # if not hasattr(variable, 'type_'):
                #     msg = 'Workflow as an untyped input variable: {}'
                #     raise WorkflowError(msg.format(variable.name))
                input_variables.append(variable)

        Block.__init__(self, input_variables, [output], name=name)
        self.output = self.outputs[0]

    def _data_hash(self):
        output_hash = self.variable_indices(self.outputs[0])
        if not isinstance(output_hash, int):
            output_hash = sum(output_hash)

        base_hash = len(self.blocks) \
            + 11 * len(self.pipes) \
            + output_hash
        block_hash = int(sum([b.equivalent_hash() for b in self.blocks])
                         % 10e5)
        return base_hash + block_hash

    def _data_eq(self, other_workflow):
        # TODO: implement imposed_variable_values in equality
        if hash(self) != hash(other_workflow):
            return False

        if not Block.equivalent(self, other_workflow):
            return False

        # TODO: temp , reuse graph!!!!
        for block1, block2 in zip(self.blocks, other_workflow.blocks):
            if not block1.equivalent(block2):
                return False
        return True

    def is_valid(self):

        return True

    def __deepcopy__(self, memo=None):
        if memo is None:
            memo = {}

        blocks = [b.__deepcopy__() for b in self.blocks]
        copied_workflow = Workflow(
            blocks=blocks, pipes=[], output=None,
            name=self.name
        )

        pipes = []
        for pipe in self.pipes:
            input_index = self.variable_indices(pipe.input_variable)

            if isinstance(input_index, int):
                pipe_input = pipe.input_variable.copy()
            elif isinstance(input_index, tuple):
                pipe_input = copied_workflow.variable_from_index(input_index)
            else:
                raise ValueError("Could not find variable at index {}".format(input_index))

            output_index = self.variable_indices(pipe.output_variable)
            pipe_output = copied_workflow.variable_from_index(output_index)
            # print(input_index, output_index)
            copied_pipe = Pipe(pipe_input, pipe_output)
            # memo[pipe] = copied_pipe

            pipes.append(copied_pipe)

        # copied_workflow.pipes = pipes

        output = copied_workflow.variable_from_index(self.variable_indices(self.output))
        # copied_workflow.output = output

        imposed_variable_values = {}
        for variable, value in self.imposed_variable_values.items():
            new_variable = copied_workflow.variable_from_index(self.variable_indices(variable))
            imposed_variable_values[new_variable] = value
        # copied_workflow.imposed_variable_values = imposed_variable_values

        copied_workflow = Workflow(
            blocks=blocks, pipes=pipes, output=output,
            imposed_variable_values=imposed_variable_values,
            name=self.name
        )
        return copied_workflow

    def _displays(self) -> List[JsonSerializable]:
        display_object = DisplayObject(type_='workflow', data=self.to_dict())
        displays = [display_object.to_dict()]
        return displays

    @property
    def _method_jsonschemas(self):
        jsonschemas = {'run': deepcopy(JSONSCHEMA_HEADER)}
        jsonschemas['run'].update({
            'classes': ['dessia_common.workflow.Workflow']
        })
        properties_dict = jsonschemas['run']['properties']
        required_inputs = []
        for i, input_ in enumerate(self.inputs):
            current_dict = {}
            annotation = (str(i), input_.type_)
            if input_ in self.nonblock_variables:
                title = input_.name
            else:
                input_block = self.block_from_variable(input_)
                if input_block.name:
                    name = input_block.name + ' - ' + input_.name
                    title = prettyname(name)
                else:
                    title = prettyname(input_.name)

            annotation_jsonschema = jsonschema_from_annotation(
                annotation=annotation, title=title,
                order=i + 1, jsonschema_element=current_dict,
            )
            # Order is i+1 because of name that is at 0
            current_dict.update(annotation_jsonschema[str(i)])
            if not input_.has_default_value:
                required_inputs.append(str(i))
            else:
                dict_ = set_default_value(
                    jsonschema_element=current_dict, key=str(i),
                    default_value=input_.default_value
                )
                current_dict.update(dict_)
            properties_dict[str(i)] = current_dict[str(i)]
        properties_dict[str(len(self.inputs) + 1)] = {
            'type': 'string', 'title': 'WorkflowRun Name', 'editable': True,
            'order': 0, 'default_value': '', 'python_typing': 'builtins.str'
        }
        jsonschemas['run']['required'] = required_inputs
        jsonschemas['run']['method'] = True
        jsonschemas['run']['python_typing'] = serialize_typing(MethodType)
        return jsonschemas

    def to_dict(self, memo=None, use_pointers=True, path='#'):
        if memo is None:
            memo = {}

        self.refresh_blocks_positions()
        dict_ = Block.to_dict(self)
        blocks = [b.to_dict() for b in self.blocks]
        pipes = []
        for pipe in self.pipes:
            pipes.append((self.variable_indices(pipe.input_variable),
                          self.variable_indices(pipe.output_variable)))

        dict_.update({'blocks': blocks, 'pipes': pipes,
                      'output': self.variable_indices(self.outputs[0]),
                      'nonblock_variables': [v.to_dict()
                                             for v in self.nonblock_variables],
                      'package_mix': self.package_mix()})

        # imposed_variables = []
        imposed_variable_values = {}
        for variable, value in self.imposed_variable_values.items():
            var_index = self.variable_indices(variable)
            # imposed_variables.append(var_index)

            if use_pointers:
                path_value = '{}/{}'.format(path, var_index)
                ser_value = serialize_with_pointers(value, memo=memo, path=path_value)

            else:
                ser_value = serialize(value)
            imposed_variable_values[var_index] = ser_value
            # if hasattr(value, 'to_dict'):
            #     imposed_variable_values.append(value.to_dict(memopath=''))
            # else:
            #     imposed_variable_values.append(value)

        # dict_['imposed_variables'] = imposed_variables
        dict_['imposed_variable_values'] = imposed_variable_values
        return dict_

    @classmethod
    def dict_to_object(cls, dict_: JsonSerializable) -> 'Workflow':
        blocks = [DessiaObject.dict_to_object(d) for d in dict_['blocks']]
        if 'nonblock_variables' in dict_:
            nonblock_variables = [dict_to_object(d)
                                  for d in dict_['nonblock_variables']]
        else:
            nonblock_variables = []

        pipes = []
        for source, target in dict_['pipes']:
            if isinstance(source, int):
                variable1 = nonblock_variables[source]
            else:
                ib1, _, ip1 = source
                variable1 = blocks[ib1].outputs[ip1]

            if isinstance(target, int):
                variable2 = nonblock_variables[target]
            else:
                ib2, _, ip2 = target
                variable2 = blocks[ib2].inputs[ip2]

            pipes.append(Pipe(variable1, variable2))

        output = blocks[dict_['output'][0]].outputs[dict_['output'][2]]
        temp_workflow = cls(blocks=blocks, pipes=pipes, output=output)

        if 'imposed_variable_values' in dict_ and 'imposed_variables' in dict_:
            # Legacy support of double list
            imposed_variable_values = {}
            iterator = zip(dict_['imposed_variables'],
                           dict_['imposed_variable_values'])
            for variable_index, serialized_value in iterator:
                value = deserialize(serialized_value)
                variable = temp_workflow.variable_from_index(variable_index)
                # if type(variable_index) == int:
                #     variable = nonblock_variables[variable_index]
                # else:
                #     iblock, side, iport = variable_index
                #     if side:
                #         variable = blocks[iblock].outputs[iport]
                #     else:
                #         variable = blocks[iblock].inputs[iport]

                imposed_variable_values[variable] = value
        elif 'imposed_variable_values' in dict_:
            # New format with a dict
            imposed_variable_values = {}
            for variable_index, serialized_value in dict_['imposed_variable_values']:
                value = deserialize(serialized_value)
                variable = temp_workflow.variable_from_index(variable_index)
                imposed_variable_values[variable] = value

        else:
            imposed_variable_values = None
        return cls(blocks=blocks, pipes=pipes, output=output,
                   imposed_variable_values=imposed_variable_values,
                   name=dict_['name'])

    def dict_to_arguments(self, dict_: JsonSerializable, method: str):
        if method in self._allowed_methods:
            arguments_values = {}
            for i, input_ in enumerate(self.inputs):
                has_default = input_.has_default_value
                if not has_default or (has_default and str(i) in dict_):
                    value = dict_[str(i)]
                    deserialized_value = deserialize_argument(
                        type_=input_.type_, argument=value
                    )
                    arguments_values[i] = deserialized_value

            name_index = len(self.inputs) + 1
            if str(name_index) in dict_:
                name = dict_[str(name_index)]
            else:
                name = None
            arguments = {'input_values': arguments_values, 'name': name}
            return arguments
        msg = 'Method {} not in Workflow allowed methods'
        raise NotImplementedError(msg.format(method))

    def method_dict(self, method_name: str = None,
                    method_jsonschema: Any = None):
        return {}

    def variable_from_index(self, index: Union[int, Tuple[int, int, int]]):
        """
        Index elements are, in order :
        - Block index : int
        - int representin port side (0: input, 1: output)
        - Port index : int
        """
        if isinstance(index, int):
            variable = self.nonblock_variables[index]
        else:
            if not index[1]:
                variable = self.blocks[index[0]].inputs[index[2]]
            else:
                variable = self.blocks[index[0]].outputs[index[2]]
        return variable

    def _get_graph(self):
        if not self._utd_graph:
            self._cached_graph = self._graph()
            self._utd_graph = True
        return self._cached_graph

    graph = property(_get_graph)

    def _graph(self):
        graph = nx.DiGraph()
        graph.add_nodes_from(self.variables)
        graph.add_nodes_from(self.blocks)
        for block in self.blocks:
            for input_parameter in block.inputs:
                graph.add_edge(input_parameter, block)
            for output_parameter in block.outputs:
                graph.add_edge(block, output_parameter)

        for pipe in self.pipes:
            graph.add_edge(pipe.input_variable, pipe.output_variable)
        return graph

    def variable_indices(self, variable):
        for iblock, block in enumerate(self.blocks):
            if variable in block.inputs:
                ib1 = iblock
                ti1 = 0
                iv1 = block.inputs.index(variable)
                return ib1, ti1, iv1
            if variable in block.outputs:
                ib1 = iblock
                ti1 = 1
                iv1 = block.outputs.index(variable)
                return ib1, ti1, iv1

        upstream_variable = self.get_upstream_nbv(variable)
        # Free variable not attached to block
        if upstream_variable in self.nonblock_variables:
            # If an upstream nbv is found, get its index
            return self.nonblock_variables.index(upstream_variable)

        msg = 'Something is wrong with variable {}'.format(variable.name)
        raise WorkflowError(msg)

    def block_from_variable(self, variable):
        iblock, _, _ = self.variable_indices(variable)
        return self.blocks[iblock]

    def output_disconnected_elements(self):
        disconnected_elements = []
        ancestors = nx.ancestors(self.graph, self.output)
        for block in self.blocks:
            if block not in ancestors:
                disconnected_elements.append(block)

        for variable in self.nonblock_variables:
            if variable not in ancestors:
                disconnected_elements.append(variable)
        return disconnected_elements

    def index(self, variable):
        warnings.warn(
            "index method is deprecated, use input_index instead",
            DeprecationWarning
        )
        index = self.input_index(variable)
        return index

    def input_index(self, variable: VariableTypes) -> int:
        upstream_variable = self.get_upstream_nbv(variable)
        return self.inputs.index(upstream_variable)

    def variable_index(self, variable: VariableTypes) -> int:
        return self.variables.index(variable)

    def get_upstream_nbv(self, variable: VariableTypes) -> VariableTypes:
        """
        If given variable has an upstream nonblock_variable, return it
        otherwise return given variable itself
        """
        if not self.nonblock_variables:
            return variable
        incoming_pipes = [p for p in self.pipes
                          if p.output_variable == variable]
        if incoming_pipes:
            # Inputs can only be connected to one pipe
            incoming_pipe = incoming_pipes[0]
            if incoming_pipe.input_variable in self.nonblock_variables:
                return incoming_pipe.input_variable
        return variable

    def layout(self, min_horizontal_spacing=300, min_vertical_spacing=200,
               max_height=800, max_length=1500):
        coordinates = {}
        elements_by_distance = {}
        for element in self.blocks + self.nonblock_variables:
            distances = []
            paths = nx.all_simple_paths(self.graph, element, self.outputs[0])
            for path in paths:
                distance = 1
                for path_element in path[1:-1]:
                    if path_element in self.blocks:
                        distance += 1
                    elif path_element in self.nonblock_variables:
                        distance += 1
                distances.append(distance)
            try:
                distance = max(distances)
            except ValueError:
                distance = 3
            if distance in elements_by_distance:
                elements_by_distance[distance].append(element)
            else:
                elements_by_distance[distance] = [element]

        if len(elements_by_distance) != 0:
            max_distance = max(elements_by_distance.keys())
        else:
            max_distance = 3  # TODO: this is an awfull quick fix

        horizontal_spacing = max(min_horizontal_spacing,
                                 max_length / max_distance)

        for i, distance in enumerate(
                sorted(elements_by_distance.keys())[::-1]):
            n = len(elements_by_distance[distance])
            vertical_spacing = min(min_vertical_spacing, max_height / n)
            horizontal_anchor_size = max_distance
            for j, element in enumerate(elements_by_distance[distance]):
                coordinates[element] = (i * horizontal_spacing,
                                        (j + 0.5) * vertical_spacing)
        return coordinates

    def refresh_blocks_positions(self):
        coordinates = self.layout()
        for i, block in enumerate(self.blocks):
            block.position = coordinates[block]
        for i, nonblock in enumerate(self.nonblock_variables):
            nonblock.position = coordinates[nonblock]

    def plot_graph(self):

        pos = nx.kamada_kawai_layout(self.graph)
        nx.draw_networkx_nodes(self.graph, pos, self.blocks,
                               node_shape='s', node_color='grey')
        nx.draw_networkx_nodes(self.graph, pos, self.variables, node_color='b')
        nx.draw_networkx_nodes(self.graph, pos, self.inputs, node_color='g')
        nx.draw_networkx_nodes(self.graph, pos, self.outputs, node_color='r')
        nx.draw_networkx_edges(self.graph, pos)

        labels = {}  # b: b.function.__name__ for b in self.block}
        for block in self.blocks:
            labels[block] = block.__class__.__name__
            for variable in self.variables:
                labels[variable] = variable.name
        nx.draw_networkx_labels(self.graph, pos, labels)

    def run(self, input_values, verbose=False,
            progress_callback=lambda x: None,
            name=None):
        log = ''

        state = self.start_run(input_values)
        state.activate_inputs(check_all_inputs=True)

        start_time = time.time()

        log_msg = 'Starting workflow run at {}'
        log_line = log_msg.format(time.strftime('%d/%m/%Y %H:%M:%S UTC',
                                                time.gmtime(start_time)))
        log += (log_line + '\n')
        if verbose:
            print(log_line)

        state.continue_run(progress_callback=progress_callback)

        end_time = time.time()
        log_line = 'Workflow terminated in {} s'.format(end_time - start_time)

        log += log_line + '\n'
        if verbose:
            print(log_line)

        state.output_value = state.values[self.outputs[0]]

        if not name:
            name = self.name + ' run'
        return state.to_workflow_run(name=name)

    def start_run(self, input_values):
        activated_items = {p: False for p in self.pipes}
        activated_items.update({v: False for v in self.variables})
        activated_items.update({b: False for b in self.blocks})

        values = {}
        variables_values = {}
        return WorkflowState(self, input_values, activated_items, values,
                             variables_values, start_time=time.time())

    # def manual_run(self, input_values, name:str=''):
    #     log = ''

    #     start_time = time.time()

    #     log_msg = 'Starting manual workflow run at {}'
    #     log_line = log_msg.format(time.strftime('%d/%m/%Y %H:%M:%S UTC',
    #                                             time.gmtime(start_time)))
    #     log += (log_line + '\n')

    #     variable_values = {}
    #     for index_input, value in input_values.items():
    #         variable_values[self.variable_indices(self.inputs[index_input])] = value

    #     return ManualWorkflowRun(workflow=self, input_values=input_values,
    #                              output_value=None,
    #                              variables_values=variable_values,
    #                              evaluated_blocks = [],
    #                              log=log, name=name)

    def mxgraph_data(self):
        nodes = []
        for block in self.blocks:
            nodes.append({'name': block.__class__.__name__,
                          'inputs': [{'name': i.name,
                                      'workflow_input': i in self.inputs,
                                      'has_default': hasattr(block, 'default')}
                                     for i in block.inputs],
                          'outputs': [{'name': o.name,
                                       'workflow_output': o in self.outputs}
                                      for o in block.outputs]})

        edges = []
        for pipe in self.pipes:
            edges.append((self.variable_indices(pipe.input_variable),
                          self.variable_indices(pipe.output_variable)))

        return nodes, edges

    def jointjs_data(self):
        coordinates = self.layout()
        blocks = []
        for block in self.blocks:
            # TOCHECK Is it necessary to add is_workflow_input/output
            #  for outputs/inputs ??
            block_data = block.jointjs_data()
            inputs = [{'name': i.name,
                       'is_workflow_input': i in self.inputs,
                       'has_default_value': hasattr(i, 'default_value')}
                      for i in block.inputs]
            outputs = [{'name': o.name,
                        'is_workflow_output': o in self.outputs}
                       for o in block.outputs]
            block_data.update({'inputs': inputs,
                               'outputs': outputs,
                               'position': coordinates[block]})
            blocks.append(block_data)

        nonblock_variables = []
        for variable in self.nonblock_variables:
            is_input = variable in self.inputs
            nonblock_variables.append({'name': variable.name,
                                       'is_workflow_input': is_input,
                                       'position': coordinates[variable]})

        edges = []
        for pipe in self.pipes:
            input_index = self.variable_indices(pipe.input_variable)
            if isinstance(input_index, int):
                node1 = input_index
            else:
                ib1, is1, ip1 = input_index
                if is1:
                    block = self.blocks[ib1]
                    ip1 += len(block.inputs)

                node1 = [ib1, ip1]

            output_index = self.variable_indices(pipe.output_variable)
            if isinstance(output_index, int):
                node2 = output_index
            else:
                ib2, is2, ip2 = output_index
                if is2:
                    block = self.blocks[ib2]
                    ip2 += len(block.inputs)

                node2 = [ib2, ip2]

            edges.append([node1, node2])

        data = Block.jointjs_data(self)
        data.update({'blocks': blocks,
                     'nonblock_variables': nonblock_variables,
                     'edges': edges})
        return data

    def plot(self):
        self.plot_jointjs(warn=False)

    def plot_jointjs(self, warn: bool = True):
        if warn:
            warnings.warn("Directly calling plot_jointjs is deprecated.\n"
                          "Please use plot instead.")
        data = json.dumps(self.jointjs_data())
        rendered_template = workflow_template.substitute(workflow_data=data)

        temp_file = tempfile.mkstemp(suffix='.html')[1]
        with open(temp_file, 'wb') as file:
            file.write(rendered_template.encode('utf-8'))

        webbrowser.open('file://' + temp_file)

    def is_valid(self):
        # Checking types of each end of pipes
        for pipe in self.pipes:
            if hasattr(pipe.input_variable, 'type_') \
                    and hasattr(pipe.output_variable, 'type_'):
                type1 = pipe.input_variable.type_
                type2 = pipe.output_variable.type_
                if type1 != type2:
                    try:
                        consistent = issubclass(pipe.input_variable.type_,
                                                pipe.output_variable.type_)
                    except TypeError:
                        # TODO: need of a real typing check
                        consistent = True
                        if not consistent:
                            msg = """
                            Inconsistent pipe type from pipe input {}
                            to pipe output {}: {} incompatible with {}
                            """
                            raise TypeError(msg.format(
                                pipe.input_variable.name,
                                pipe.output_variable.name,
                                pipe.input_variable.type_,
                                pipe.output_variable.type_)
                            )
        return True

    def package_mix(self):
        """

        """
        package_mix = {}
        for block in self.blocks:
            if hasattr(block, 'package_mix'):
                for package_name, fraction in block.package_mix().items():
                    if package_name in package_mix:
                        package_mix[package_name] += fraction
                    else:
                        package_mix[package_name] = fraction

        # Adimension
        fraction_sum = sum(package_mix.values())
        return {pn: f / fraction_sum for pn, f in package_mix.items()}


class WorkflowBlock(Block):
    """
    Wrapper around workflow to put it in a block of another workflow
    Even if a workflow is a block, it can't be used directly as it has
    a different behavior
    than a Block in eq and hash which is problematic to handle in dicts
    for example

    :param workflow: The WorkflowBlock's workflow
    :type workflow: Workflow
    """

    def __init__(self, workflow: Workflow,
                 name: str = ''):
        self.workflow = workflow
        # TODO: configuring port internal connections
        self.input_connections = None
        self.output_connections = None
        inputs = []
        for i, variable in enumerate(self.workflow.inputs):
            input_ = variable.copy()
            input_.name = '{} - {}'.format(name, variable.name)
            inputs.append(input_)

        outputs = [self.workflow.output.copy()]
        Block.__init__(self, inputs, outputs, name=name)

    def equivalent_hash(self):
        return hash(self.workflow)

    def equivalent(self, other):
        if not Block.equivalent(self, other):
            return False
        return self.workflow == other.workflow

    def to_dict(self):
        dict_ = Block.to_dict(self)
        dict_.update({'workflow': self.workflow.to_dict()})
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_):
        b = cls(workflow=Workflow.dict_to_object(dict_['workflow']),
                name=dict_['name'])

        return b

    def evaluate(self, values):
        args = {self.inputs.index(input_): v for input_, v in values.items()}
        workflow_run = self.workflow.run(args)
        return [workflow_run.output_value]

    def package_mix(self):
        return self.workflow.package_mix()


class WorkflowState(DessiaObject):
    _standalone_in_db = True
    _allowed_methods = ['block_evaluation', 'evaluate_next_block',
                        'evaluate_maximum_blocks', 'add_block_input_values']
    _non_serializable_attributes = ['activated_items']

    def __init__(self, workflow: Workflow, input_values, activated_items,
                 values, variables_values, start_time, output_value=None,
                 log: str = '', name: str = ''):
        self.workflow = workflow
        self.input_values = input_values
        self.output_value = output_value
        self.variables_values = variables_values
        self.values = values
        self.start_time = start_time
        self.log = log

        self.activated_items = activated_items

        DessiaObject.__init__(self, name=name)

    def to_dict(self, use_pointers: bool = True, memo=None, path: str = '#'):
        if not use_pointers:
            msg = 'WorkflowState to_dict should not' \
                  'be called with use_pointers=False'
            raise NotImplementedError(msg)
        if memo is None:
            memo = {}

        dict_ = DessiaObject.to_dict(self)

        values = {self.workflow.variable_index(i): serialize(v)
                  for i, v in self.values.items()}
        dict_.update({'values': values})
        dict_['evaluated_blocks_indices'] = [i for i, b
                                             in enumerate(self.workflow.blocks)
                                             if b in self.activated_items
                                             and self.activated_items[b]]
        dict_['evaluated_pipes_indices'] = [i for i, p
                                            in enumerate(self.workflow.pipes)
                                            if p in self.activated_items
                                            and self.activated_items[p]]
        dict_['evaluated_variables_indices'] = [
            self.workflow.variable_indices(v) for v in self.workflow.variables
            if v in self.activated_items and self.activated_items[v]
        ]
        if self.output_value is not None:
            dict_.update({
                'output_value': serialize(self.output_value),
                'output_value_type': recursive_type(self.output_value)
            })
        return dict_

    @classmethod
    def dict_to_object(cls, dict_: JsonSerializable,
                       force_generic: bool = False) -> 'WorkflowState':
        workflow = Workflow.dict_to_object(dict_['workflow'])
        if 'output_value' in dict_ and 'output_value_type' in dict_:
            type_ = dict_['output_value_type']
            value = dict_['output_value']
            output_value = recursive_instantiation(type_=type_, value=value)
        else:
            output_value = None

        values = {workflow.variables[int(i)]: deserialize(v)
                  for i, v in dict_['values'].items()}
        input_values = {int(i): deserialize(v)
                        for i, v in dict_['input_values'].items()}
        variables_values = {k: deserialize(v)
                            for k, v in dict_['variables_values'].items()}

        activated_items = {
            b: (True if i in dict_['evaluated_blocks_indices'] else False)
            for i, b in enumerate(workflow.blocks)
        }

        activated_items.update({
            p: (True if i in dict_['evaluated_pipes_indices'] else False)
            for i, p in enumerate(workflow.pipes)
        })

        var_indices = dict_['evaluated_variables_indices']
        activated_items.update({
            v: (True if workflow.variable_indices(v) in var_indices else False)
            for v in workflow.variables
        })

        return cls(workflow=workflow, input_values=input_values,
                   activated_items=activated_items, values=values,
                   variables_values=variables_values,
                   start_time=dict_['start_time'], output_value=output_value,
                   log=dict_['log'], name=dict_['name'])

    def add_input_value(self, input_index: int, value: Any):
        # TODO: Type checking?
        self.input_values[input_index] = value
        self.activate_inputs()

    def add_several_input_values(self, indices: List[int],
                                 values: Dict[str, Any]):
        for i in indices:
            self.add_input_value(input_index=i, value=values[str(i)])

    def add_block_input_values(self, block_index: int,
                               values: Dict[str, Any]):
        block = self.workflow.blocks[block_index]
        indices = [self.workflow.input_index(i) for i in block.inputs]
        self.add_several_input_values(indices=indices, values=values)

    def _displays(self) -> List[JsonSerializable]:
        data = self.to_dict()

        display_object = DisplayObject(type_='workflow_state', data=data)
        displays = [display_object.to_dict()]
        return displays

    @property
    def progress(self):
        activated_items = [b for b in self.workflow.blocks
                           if b in self.activated_items
                           and self.activated_items[b]]
        return len(activated_items) / len(self.workflow.blocks)

    def block_evaluation(self, block_index: int,
                         progress_callback=lambda x: None) -> bool:
        """
        Select a block to evaluate
        """
        block = self.workflow.blocks[block_index]

        self.activate_inputs()
        for pipe in self._activable_pipes():
            self._evaluate_pipe(pipe)

        if block in self._activable_blocks():
            self._evaluate_block(block)
            progress_callback(self.progress)
            return True
        else:
            return False

    def evaluate_next_block(self, progress_callback=lambda x: None) -> Optional[Block]:
        """
        Evaluate a block
        """
        self.activate_inputs()
        for pipe in self._activable_pipes():
            self._evaluate_pipe(pipe)

        blocks = self._activable_blocks()
        if blocks:
            block = blocks[0]
            self._evaluate_block(block)
            progress_callback(self.progress)
            return block
        else:
            return None

    def continue_run(self, progress_callback=lambda x: None):
        """
        Evaluate all possible blocks
        """
        self.activate_inputs()

        evaluated_blocks = []
        something_activated = True
        while something_activated:
            something_activated = False

            for pipe in self._activable_pipes():
                self._evaluate_pipe(pipe)
                something_activated = True

            for block in self._activable_blocks():
                evaluated_blocks.append(block)
                self._evaluate_block(block)
                progress_callback(self.progress)
                something_activated = True
        return evaluated_blocks

    def _activable_pipes(self):
        pipes = []
        for pipe in self.workflow.pipes:
            if not self.activated_items[pipe]:
                if self.activated_items[pipe.input_variable]:
                    pipes.append(pipe)
        return pipes

    def _activable_blocks(self):
        blocks = []
        for block in self.workflow.blocks:
            if not self.activated_items[block]:
                all_inputs_activated = True
                for function_input in block.inputs:
                    if not self.activated_items[function_input]:
                        all_inputs_activated = False
                        break

                if all_inputs_activated:
                    blocks.append(block)
        return blocks

    def _evaluate_pipe(self, pipe):
        self.activated_items[pipe] = True
        self.values[pipe.output_variable] = self.values[
            pipe.input_variable]
        self.activated_items[pipe.output_variable] = True

    def _evaluate_block(self, block, progress_callback=lambda x: x,
                        verbose=False):
        if verbose:
            log_line = 'Evaluating block {}'.format(block.name)
            self.log += log_line + '\n'
            if verbose:
                print(log_line)

        output_values = block.evaluate({i: self.values[i]
                                        for i in block.inputs})
        for input_ in block.inputs:
            if input_.memorize:
                indices = str(self.workflow.variable_indices(input_))  # Str is strange
                self.variables_values[indices] = self.values[input_]
        # Updating progress
        if progress_callback is not None:
            progress_callback(self.progress)

        # Unpacking result of evaluation
        output_items = zip(block.outputs, output_values)
        for output, output_value in output_items:
            if output.memorize:
                indices = str(self.workflow.variable_indices(output))
                self.variables_values[indices] = output_value
            self.values[output] = output_value
            self.activated_items[output] = True

        self.activated_items[block] = True

    def activate_inputs(self, check_all_inputs=False):
        """
        Returns if all inputs are activated
        """
        # Imposed variables values activation
        for variable, value in self.workflow.imposed_variable_values.items():
            # Type checking
            value_type_check(value, variable.type_)
            self.values[variable] = value
            self.activated_items[variable] = True

        # Input activation
        for index, variable in enumerate(self.workflow.inputs):
            if index in self.input_values:
                value = self.input_values[index]
                self.values[variable] = value
                self.activated_items[variable] = True
            elif hasattr(variable, 'default_value'):
                self.values[variable] = variable.default_value
                self.activated_items[variable] = True
            elif check_all_inputs:
                msg = 'Value {} of index {} in inputs has no value'
                if isinstance(variable, TypedVariable):
                    msg += ': should be instance of {}'.format(variable.type_)
                raise ValueError(msg.format(variable.name, index))

    def to_workflow_run(self, name=''):
        if self.progress == 1:
            return WorkflowRun(workflow=self.workflow,
                               input_values=self.input_values,
                               output_value=self.output_value,
                               variables_values=self.variables_values,
                               start_time=self.start_time, end_time=time.time(),
                               log=self.log, name=name)
        else:
            raise ValueError('Workflow not completed')


class WorkflowRun(DessiaObject):
    _standalone_in_db = True
    _allowed_methods = ['run_again']
    _eq_is_data_eq = True
    _jsonschema = {
        "definitions": {},
        "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "object",
        "title": "WorkflowRun Base Schema",
        "required": [],
        "python_typing": 'dessia_common.workflow.Pipe',
        "standalone_in_db": True,
        "properties": {
            "workflow": {
                "type": "object", "title": "Workflow",
                "python_typing": "dessia_common.workflow.Workflow",
                "classes": ["dessia_common.workflow.Workflow"],
                "order": 0, "editable": False, "description": "Workflow"
            },
            'output_value': {
                "type": "object", "classes": "Any", "title": "Values",
                "description": "Input and output values",
                "editable": False, "order": 1, "python_typing": "Any"
            },
            'input_values': {
                'type': 'object', 'order': 2, 'editable': False,
                'title': 'Input Values', "python_typing": "Dict[str, Any]",
                'patternProperties': {
                    '.*': {
                        'type': "object",
                        'classes': 'Any'
                    }
                }
            },
            'variables_values': {
                'type': 'object', 'order': 3, 'editable': False,
                'title': 'Variables Values', "python_typing": "Dict[str, Any]",
                'patternProperties': {
                    '.*': {
                        'type': "object",
                        'classes': 'Any'
                    }
                }
            },
            'start_time': {
                "type": "number", "title": "Start Time",
                "editable": False, "python_typing": "builtins.int",
                "description": "Start time of simulation", "order": 4
            },
            'end_time': {
                "type": "number", "title": "End Time",
                "editable": False, "python_typing": "builtins.int",
                "description": "End time of simulation", "order": 5
            },
            'log': {
                "type": "string", "title": "Log",
                "editable": False, 'python_typing': 'builtins.str',
                "description": "Log", "order": 6,
            },
            "name": {
                'type': 'string', 'title': 'Name', 'editable': True,
                'order': 7, 'default_value': '',
                'python_typing': 'builtins.str'
            }
        }
    }

    def __init__(self, workflow, input_values, output_value, variables_values,
                 start_time, end_time, log, name=''):
        self.workflow = workflow
        self.input_values = input_values
        self.output_value = output_value
        self.variables_values = variables_values
        self.start_time = start_time
        self.end_time = end_time
        self.execution_time = end_time - start_time
        self.log = log

        DessiaObject.__init__(self, name=name)

    def _data_eq(self, other_workflow_run):
        # TODO : Should we add input_values and variables values in test ?
        if not data_eq(self.output_value, other_workflow_run.output_value):
            return False
        return self.workflow._data_eq(other_workflow_run.workflow)

    def _data_hash(self):
        # TODO : Should we add input_values and variables values in test ?
        if is_sequence(self.output_value):
            hash_output = list_hash(self.output_value)
        else:
            hash_output = hash(self.output_value)
        return hash(self.workflow) + int(hash_output % 10e5)

    def _displays(self) -> List[JsonSerializable]:
        d_blocks = [b for b in self.workflow.blocks if hasattr(b, 'display_')]
        sorted_d_blocks = sorted(d_blocks, key=lambda b: b.order)
        displays = self.workflow._displays()
        for block in sorted_d_blocks:
            reference_path = ''
            local_values = {}
            for i, input_ in enumerate(block.inputs):
                strindices = str(self.workflow.variable_indices(input_))
                local_values[input_] = self.variables_values[strindices]
                if i == block._displayable_input:
                    reference_path = 'variables_values/' + strindices
            display = block.display_(local_values=local_values,
                                     reference_path=reference_path)
            displays.extend(display)
        if isinstance(self.output_value, DessiaObject):
            displays.extend(self.output_value._displays(
                reference_path='output_value'
            ))
        return displays

    # @classmethod
    # def dict_to_object(cls, dict_):
    #     workflow = Workflow.dict_to_object(dict_['workflow'])
    #     if 'output_value' in dict_ and 'output_value_type' in dict_:
    #         type_ = dict_['output_value_type']
    #         value = dict_['output_value']
    #         output_value = recursive_instantiation(type_=type_, value=value)
    #     else:
    #         output_value = None

    #     input_values = {int(i): deserialize(v)
    #                     for i, v in dict_['input_values'].items()}
    #     variables_values = {k: deserialize(v)
    #                         for k, v in dict_['variables_values'].items()}
    #     return cls(workflow=workflow, output_value=output_value,
    #                input_values=input_values,
    #                variables_values=variables_values,
    #                start_time=dict_['start_time'], end_time=dict_['end_time'],
    #                log=dict_['log'], name=dict_['name'])

    def to_dict(self, use_pointers: bool = True, memo=None, path: str = '#'):
        input_values = {}

        if not use_pointers:
            raise NotImplementedError('WorkflowRun to_dict should not be called with use_pointers=False')

        if memo is None:
            memo = {}

        for i, v in self.input_values.items():
            serialized_v, memo = serialize_with_pointers(v, memo, path='#/input_values/{}'.format(i))
            input_values[i] = serialized_v

        variables_values = {}
        for k, v in self.variables_values.items():
            serialized_v, memo = serialize_with_pointers(v, memo,
                                                         path='#/variables_values/{}'.format(k))
            variables_values[k] = serialized_v

        # variables_values = {k: serialize(v)
        #                     for k, v in self.variables_values.items()}
        workflow = self.workflow.to_dict(path='#/workflow', memo=memo)

        dict_ = DessiaObject.base_dict(self)
        dict_.update({'workflow': workflow,
                      'input_values': input_values,
                      'variables_values': variables_values,
                      'start_time': self.start_time, 'end_time': self.end_time,
                      'execution_time': self.execution_time, 'log': self.log})

        if self.output_value is not None:
            serialized_output, _ = serialize_with_pointers(self.output_value, memo, path='#/output_value')
            dict_.update({
                'output_value': serialized_output,
                'output_value_type': recursive_type(self.output_value)
            })

        return dict_

    def dict_to_arguments(self, dict_: JsonSerializable, method: str):
        if method in self._allowed_methods:
            return self.workflow.dict_to_arguments(dict_=dict_, method='run')
        msg = 'Method {} not in WorkflowRun allowed methods'
        raise NotImplementedError(msg.format(method))

    def method_dict(self, method_name: str = None,
                    method_jsonschema: Any = None):
        if method_name is not None and method_name == 'run_again' \
                and method_jsonschema is not None:
            dict_ = serialize_dict(self.input_values)
            for property_, value in method_jsonschema['properties'].items():
                if property_ in dict_ \
                        and 'object_id' in value and 'object_class' in value:
                    # TODO : Check. this is probably useless as we are not dealing with default values here
                    dict_[property_] = value
            return dict_
        # TODO Check this result. Might raise an error
        return DessiaObject.method_dict(self, method_name=method_name,
                                        method_jsonschema=method_jsonschema)

    def run_again(self, input_values, progress_callback=None, name=None):
        workflow_run = self.workflow.run(input_values=input_values,
                                         verbose=False,
                                         progress_callback=progress_callback,
                                         name=name)
        return workflow_run

    @property
    def _method_jsonschemas(self):
        jsonschemas = self.workflow._method_jsonschemas
        jsonschemas['run_again'] = jsonschemas.pop('run')
        workflow_run_class = "dessia_common.workflow.WorkflowRun"
        jsonschemas['run_again']['classes'] = [workflow_run_class]
        return jsonschemas


def set_inputs_from_function(method, inputs=None):
    """

    """
    if inputs is None:
        inputs = []
    args_specs = inspect.getfullargspec(method)
    nargs = len(args_specs.args) - 1

    if args_specs.defaults is not None:
        ndefault_args = len(args_specs.defaults)
    else:
        ndefault_args = 0

    for iarg, argument in enumerate(args_specs.args[1:]):
        if argument not in ['self', 'progress_callback']:
            try:
                annotations = get_type_hints(method)
                type_ = type_from_annotation(annotations[argument],
                                             module=method.__module__)
            except KeyError:
                msg = 'Argument {} of method/function {} has no typing'
                raise UntypedArgumentError(
                    msg.format(argument, method.__name__))
            if iarg >= nargs - ndefault_args:
                default = args_specs.defaults[ndefault_args - nargs + iarg]
                input_ = TypedVariableWithDefaultValue(type_=type_,
                                                       default_value=default,
                                                       name=argument)
                inputs.append(input_)

            else:
                inputs.append(TypedVariable(type_=type_, name=argument))
    return inputs


def value_type_check(value, type_):
    try:  # TODO: Subscripted generics cannot be used...
        if not isinstance(value, type_):
            return False
    except TypeError:
        pass

    return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 19 15:56:12 2020

@author: jezequel
"""

from io import StringIO
import math
from typing import List, Dict, Any
import numpy as np
import pandas as pd
from dessia_common import DessiaObject, Parameter, is_bounded, DessiaFilter
from scipy.optimize import minimize
from pyDOE import lhs


class ParetoSettings(DessiaObject):
    """
    :param coeff_names: A dictionary containing the name of a variable as key and \
    a boolean for minimization as value.
    :type coeff_names: Dict[str, bool]
    :param enabled: List of strings representing ordered variables names.
    :type enabled: bool
    :param name: The name of the block.
    :type name: str
    """
    _generic_eq = True
    _ordered_attributes = ['name', 'enabled', 'minimized_attributes']

    def __init__(self, minimized_attributes: Dict[str, bool],
                 enabled: bool = True, name: str = ''):
        self.enabled = enabled
        self.minimized_attributes = minimized_attributes

        DessiaObject.__init__(self, name=name)


class ObjectiveSettings(DessiaObject):
    _generic_eq = True
    _ordered_attributes = ['name', 'enabled', 'n_near_values']

    def __init__(self, n_near_values: int = 4,
                 enabled: bool = True, name: str = ''):
        self.n_near_values = n_near_values
        self.enabled = enabled

        DessiaObject.__init__(self, name=name)


class Objective(DessiaObject):
    """
    Defines an objective function

    :param coeff_names: List of strings representing ordered variables names.
    :type coeff_names: [str]
    :param coeff_values: List of floats representing coefficients.
    :type coeff_values: [float]

    Order is kept. It means that coefficients are applied to the variables
    in the same order as they are defined.
    """
    _generic_eq = True
    _standalone_in_db = False
    _ordered_attributes = ['name', 'settings', 'coefficients']
    _non_serializable_attributes = ['coeff_names']

    def __init__(self, coefficients: Dict[str, float],
                 directions: Dict[str, bool],
                 settings: ObjectiveSettings, name: str = ''):
        for variable in coefficients:
            if variable not in directions:
                msg = "Coefficient for variable {}".format(variable)
                msg += " was found but no direction is specified."
                msg += " Add {} to directions dict.".format(variable)
                raise KeyError(msg)
        self.coefficients = coefficients
        self.directions = directions
        self.coeff_names = list(coefficients.keys())
        self.settings = settings

        DessiaObject.__init__(self, name=name)

    def apply_individual(self, values: Dict[str, float]) -> float:
        rating = 0
        for variable, value in values.items():
            coefficient = self.coefficients[variable]
            rating += coefficient * value
        return rating

    @classmethod
    def coefficients_from_angles(cls, angles: List[float]) -> List[float]:
        """
        compute coefficients from n-1 angles
        """
        n = len(angles) + 1
        matrix = np.identity(n)
        for i, angle in enumerate(angles):
            matrix_i = np.identity(n)
            matrix_i[i, i] = math.cos(angle)
            matrix_i[i + 1, i] = -math.sin(angle)
            matrix_i[i, i + 1] = math.sin(angle)
            matrix_i[i + 1, i + 1] = math.cos(angle)
            matrix = np.dot(matrix, matrix_i)
        x = np.zeros(n)
        x[0] = 1
        signed = np.dot(matrix.T, x).tolist()
        unsigned = [v for v in signed]
        return unsigned

    @classmethod
    def from_angles(cls, angles: List[float], variables: List[str],
                    directions: Dict[str, bool],
                    settings: ObjectiveSettings = None,
                    name="Generated from angles") -> 'Objective':
        if not isinstance(angles, list) and not isinstance(angles, np.ndarray):
            angles = [angles]
        generated_coefficients = cls.coefficients_from_angles(angles=angles)
        coefficients = {}
        for i, var in enumerate(variables):
            coeff = generated_coefficients[i]
            if (directions[var] and coeff < 0)\
                    or (not directions[var] and coeff >= 0):
                coefficients[var] = -coeff
            else:
                coefficients[var] = coeff

        if settings is None:
            settings = ObjectiveSettings()

        return Objective(coefficients=coefficients, directions=directions,
                         settings=settings, name=name)


class Catalog(DessiaObject):
    """
    Defines a Catalog object that gathers a collection of VectoredObjects

    TODO Update Docstring
    :param objectives: List of objectives to apply to
                       catalog vectored objects
    :type objectives: [Objective]
    # :param objects: List of vectored objects.
    # :type objects: [VectoredObject]
    :param choice_variables: List of string. List of variable names
                             that represent choice arguments
    :type choice_variables: [str]
    :param name: Name of the catalog
    :type name: str

    """
    _generic_eq = True
    _standalone_in_db = True
    _ordered_attributes = ['name', 'pareto_settings', 'objectives']
    _non_editable_attributes = ['array', 'variables', 'choice_variables',
                                'generated_best_objectives']
    # _export_formats = ['csv']
    _allowed_methods = ['find_best_objective']
    _whitelist_attributes = ['variables']

    def __init__(self, array: List[List[float]], variables: List[str],
                 pareto_settings: ParetoSettings = None,
                 objectives: List[Objective] = None,
                 choice_variables: List[str] = None,
                 name: str = ''):
        DessiaObject.__init__(self, name=name)

        self.array = array
        self.variables = variables
        if choice_variables is None:
            self.choice_variables = variables
        else:
            self.choice_variables = choice_variables

        self.pareto_settings = pareto_settings

        if objectives is None:
            self.objectives = []
        else:
            self.objectives = objectives

        self.generated_best_objectives = 0

    @classmethod
    def concatenate(cls, catalogs: List['Catalog'],
                    pareto_settings: ParetoSettings = None,
                    objectives: List[Objective] = None,
                    choice_variables: List[str] = None,
                    name: str = '') -> 'Catalog':
        varsets = [set(c.variables) for c in catalogs]
        var_intersection = list(set.intersection(*varsets))

        if choice_variables is not None:
            choice_variables = [var for var in choice_variables
                                if var in var_intersection]
        else:
            choice_variables = var_intersection

        array_intersection = []
        for cat in catalogs:
            indices = [cat.get_variable_index(v) for v in var_intersection]
            for line in cat.array:
                line_intersection = [v for j, v in enumerate(line)
                                     if j in indices]
                array_intersection.append(line_intersection)

        if pareto_settings is None:
            min_attrs = {var: True for var in choice_variables}
            pareto_settings = ParetoSettings(minimized_attributes=min_attrs,
                                             enabled=False)

        catalog = cls(array=array_intersection, variables=var_intersection,
                      pareto_settings=pareto_settings, objectives=objectives,
                      choice_variables=choice_variables, name=name)
        return catalog

    # def __getitem__(self, item):
    #     print('Getitem', item)
    #     return self.array[item]
    #
    # def __getattr__(self, item):
    #     if isinstance(item, (list, tuple)):
    #         return enhanced_deep_attr(self, item)w
    #     elif isinstance(item, str) and '.' in item:
    #         sequence = item.split('.')
    #         healed_sequence = []
    #         for i, attr in enumerate(sequence):
    #             try:
    #                 healed_sequence.append(int(attr))
    #             except ValueError:
    #                 if attr in self.variables:
    #                     healed_sequence.append(self.get_variable_index(attr))
    #                 else:
    #                     healed_sequence.append(attr)
    #         print(healed_sequence)
    #         return enhanced_deep_attr(self, healed_sequence)
    #     return self.__getattribute__(item)

    def generate_multiplot(self, values: Dict[str, Any] = None):
        # TOCHECK Avoid circular imports
        import plot_data
        from plot_data.colors import BLACK, LIGHTBLUE, LIGHTGREY, BLUE

        if values is None:
            values = []
            for i, line in enumerate(self.array):
                value = {}
                for variable in self.variables:
                    value[variable] = self.get_value_by_name(line, variable)
                # for objective_name, ratings in objective_ratings.items():
                #     value[objective_name] = ratings[i]
                values.append(value)

        fontsize = 12
        first_vars = self.variables[:2]
        values2d = [{key: val[key]} for key in first_vars for val in values]
        rgbs = [[192, 11, 11], [14, 192, 11], [11, 11, 192]]

        tooltip = plot_data.Tooltip(attributes=self.variables,
                                    name='Tooltip')

        scatterplot = plot_data.Scatter(axis=plot_data.Axis(), tooltip=tooltip,
                                        x_variable=first_vars[0],
                                        y_variable=first_vars[1],
                                        elements=values2d, name='Scatter Plot')

        parallelplot = plot_data.ParallelPlot(disposition='horizontal',
                                              axes=self.variables,
                                              rgbs=rgbs, elements=values)
        objects = [scatterplot, parallelplot]
        sizes = [plot_data.Window(width=560, height=300),
                 plot_data.Window(width=560, height=300)]
        coords = [(0, 0), (0, 300)]
        multiplot = plot_data.MultiplePlots(plots=objects, elements=values,
                                            sizes=sizes, coords=coords,
                                            name='Results plot')
        return multiplot

    def filter_(self, filters: List[DessiaFilter]):
        def apply_filters(line):
            bounded = True
            i = 0
            while bounded and i < len(filters):
                filter_ = filters[i]
                variable = filter_.attribute
                value = line[self.get_variable_index(variable)]
                bounded = is_bounded(filter_, value)
                i += 1
            return bounded

        filtered_array = list(filter(apply_filters, self.array))
        return filtered_array

    @classmethod
    def from_csv(cls, file: StringIO, end: int = None, remove_duplicates: bool = False):
        """
        Generates MBSEs from given .csv file.
        """
        array = np.genfromtxt(file, dtype=None, delimiter=',',
                              names=True, encoding=None)
        variables = [v for v in array.dtype.fields.keys()]
        lines = []
        for i, line in enumerate(array):
            if end is not None and i >= end:
                break
            if not remove_duplicates or (remove_duplicates
                                         and line.tolist() not in lines):
                lines.append(line.tolist())
        return cls(lines, variables)

    def export_csv(self, attribute_name: str, indices: List[int], file: str):
        """
        Exports a reduced list of objects to .csv file

        :param attribute_name: Name of the attribute
                               in which the list is stored
        :type attribute_name: str
        :param indices: List of integers that represents selected
                        indices of object in attribute_name sequence
        :type indices: [int]
        :param file: Target file
        """
        attribute = getattr(self, attribute_name)
        lines = [attribute[i] for i in indices]
        array = np.array(lines)
        data_frame = pd.DataFrame(array, columns=self.variables)
        data_frame.to_csv(file, index=False)

    def parameters(self, variables: List[str]) -> List[Parameter]:
        """
        Computes Parameter objects from catalog structural data

        :param variables: List of string. Names of arguments of which
                         it should create a parameter.
        :type variables: [string]

        :return: List of Parameter objects
        """
        parameters = []
        for variable in variables:
            values = self.get_values(variable)
            parameters.append(Parameter(lower_bound=min(values),
                                        upper_bound=max(values),
                                        name=variable))
        return parameters

    def get_variable_index(self, name: str) -> int:
        return self.variables.index(name)

    def get_values(self, variable: str) -> List:
        values = [self.get_value_by_name(line, variable)
                  for line in self.array]
        return values

    def get_value_by_name(self, line: List, name: str) -> float:
        j = self.get_variable_index(name)
        value = line[j]
        return value

    @property
    def pareto_is_enabled(self) -> bool:
        pareto = self.pareto_settings
        return pareto is not None and pareto.enabled

    def build_costs(self, pareto_settings: ParetoSettings):
        """
        Build list of costs that are used to compute Pareto frontier.

        The cost of an attribute that is to be minimized is,
        for each object of catalog, its value minus the lower bound
        of its values in the whole dataset.
        On the contrary, the cost of an attribute that is to be
        maximised, is the upper_bound of the dataset for this parameter
        minus the value of the attribute of each object of the catalog.

        For a Pareto frontier of dimensions n_costs, each vectored
        bject of the catalog (n_points vectored objects in the catalog)
        will give a numpy array of dimensions (,n_costs)

        All put together build_costs method results in a numpy array :

        :return: A(n_points, n_costs)
        """
        params = list(pareto_settings.minimized_attributes.keys())
        pareto_parameters = self.parameters(params)
        costs = np.zeros((len(self.array), len(pareto_parameters)))
        for i, line in enumerate(self.array):
            for j, parameter in enumerate(pareto_parameters):
                if pareto_settings.minimized_attributes[parameter.name]:
                    value = self.get_value_by_name(line, parameter.name)\
                            - parameter.lower_bound
                else:
                    value = parameter.upper_bound\
                            - self.get_value_by_name(line, parameter.name)
                costs[(i, j)] = value
        return costs

    @classmethod
    def random_2d(cls, bounds: Dict[str, List[float]], threshold: float,
                  end: float = 500, name='Random Set'):
        """
        This method is for dev purpose. It can be removed if needed
        """
        array = []
        variables = list(bounds.keys())
        pareto_settings = ParetoSettings({v: True for v in variables},
                                         enabled=True)
        objectives = []
        while len(array) <= end:
            line = [(bounds[v][1] - bounds[v][0]) * np.random.rand() + bounds[v][0]
                    for v in variables]
            if line[0] * line[1] >= threshold:
                array.append(line)
        return cls(array=array, variables=variables,
                   pareto_settings=pareto_settings, objectives=objectives,
                   choice_variables=variables, name=name)

    def handle_objective(self, objective: Objective) -> List[float]:
        ratings = []
        for line in self.array:
            values = {variable: self.get_value_by_name(line, variable)
                      for variable in objective.coefficients}
            rating = objective.apply_individual(values)
            ratings.append(rating)
        return ratings

    def find_best_objective(self, values: Dict[str, float],
                            minimized: Dict[str, bool],
                            settings: ObjectiveSettings = None):
        # Unordered list of variables
        variables = list(values.keys())  # !!!
        parameters = self.parameters(variables)
        names = [p.name for p in parameters]

        # Get pareto points
        pareto_settings = ParetoSettings(minimized_attributes=minimized)
        costs = self.build_costs(pareto_settings=pareto_settings)
        pareto_indices = pareto_frontier(costs=costs)
        pareto_values = [{var: self.get_value_by_name(self.array[i], var)
                          for var in variables}
                         for i, is_efficient in enumerate(pareto_indices)
                         if is_efficient]

        def apply(x):
            objective = Objective.from_angles(angles=x,
                                              variables=names,
                                              directions=minimized)
            best_on_pareto = min([objective.apply_individual(pareto_value)
                                  for pareto_value in pareto_values])
            rating = objective.apply_individual(values)
            delta = rating - best_on_pareto
            return delta

        i = 0
        success = False
        randomized_angles = lhs(len(values) - 1, 100)
        available_angles = [[angle * 2 * math.pi for angle in angles]
                            for angles in randomized_angles]
        while i < len(available_angles) and not success:
            res = minimize(fun=apply, x0=randomized_angles[i], method="Powell")
            i += 1
            if res.success:
                success = True
                best_objective = Objective.from_angles(angles=res.x.tolist(),
                                                       variables=names,
                                                       directions=minimized,
                                                       settings=settings)
                best_objective.name = "Best Coefficients"\
                                      + str(self.generated_best_objectives)
                self.generated_best_objectives += 1
                self.objectives.append(best_objective)
        if not success:
            raise ValueError("No solutions found")

    def plot_data(self):
        from plot_data.core import Tooltip, TextStyle, SurfaceStyle,\
            Scatter, ParallelPlot, PointFamily, MultiplePlots
        from plot_data.colors import GREY, LIGHTGREY, LIGHTGREEN, LIGHTBLUE

        name_column_0 = self.variables[0]
        list_name = self.get_values(name_column_0)

        list_settings = [name
                         for name in self.pareto_settings.minimized_attributes]
        list_value = [self.get_values(cv) for cv in self.choice_variables]
        if self.pareto_is_enabled:
            cost = self.build_costs(self.pareto_settings)
            p_frontier = pareto_frontier(cost)
            elements = [[], []]  # point_non_pareto, point_pareto
            for i in range(len(list_name)):
                dict_element = {name_column_0: list_name[i]}
                for k, setting in enumerate(self.choice_variables):
                    dict_element[setting] = list_value[k][i]

                if p_frontier[i]:
                    elements[1].append(dict_element)
                else:
                    elements[0].append(dict_element)

        else:
            elements = [[], []]
            for i in range(len(list_name)):
                dict_element = {name_column_0: list_name[i]}
                for k, setting in enumerate(self.choice_variables):
                    dict_element[setting] = list_value[k][i]

                elements[0].append(dict_element)

        # ScatterPlot
        attributes = self.choice_variables
        text_style = TextStyle(text_color=GREY,
                               font_size=10,
                               font_style='sans-serif')
        surface_style = SurfaceStyle(color_fill=LIGHTGREY, opacity=0.3)
        custom_tooltip = Tooltip(attributes=attributes,
                                 surface_style=surface_style,
                                 text_style=text_style,
                                 tooltip_radius=10)

        all_points = elements[0] + elements[1]

        plots = []

        # ScatterPlot
        for j in range(len(list_settings)):
            for i in range(j + 1, len(list_settings)):
                if len(plots) < 3:
                    plots.append(Scatter(tooltip=custom_tooltip,
                                         x_variable=list_settings[j],
                                         y_variable=list_settings[i],
                                         elements=all_points))

        list_index_0 = [k for k in range(len(elements[0]))]
        point_family_0 = PointFamily(LIGHTBLUE, list_index_0,
                                     name='Non pareto')

        n_pareto = len(elements[1])
        list_index_1 = [len(all_points) - i - 1 for i in range(n_pareto)]
        point_family_1 = PointFamily(LIGHTGREEN, list_index_1, name='Pareto')

        # ParallelPlot
        plots.append(ParallelPlot(elements=all_points,
                                  axes=attributes))

        return [MultiplePlots(plots=plots, elements=all_points,
                              point_families=[point_family_0, point_family_1],
                              initial_view_on=True)]


def from_csv(filename: str, end: int = None, remove_duplicates: bool = False):
    """
    Generates MBSEs from given .csv file.
    """
    array = np.genfromtxt(filename, dtype=None, delimiter=',',
                          names=True, encoding=None)
    variables = [v for v in array.dtype.fields.keys()]
    lines = []
    for i, line in enumerate(array):
        if end is not None and i >= end:
            break
        if not remove_duplicates or (remove_duplicates
                                     and line.tolist() not in lines):
            lines.append(line.tolist())
    return lines, variables


def pareto_frontier(costs):
    """
    Find the pareto-efficient points

    :param catalog: Catalog object on which to apply pareto_frontier
                    computation
    :return: A (n_points, ) boolean array, indicating whether each point
             is Pareto efficient
    """
    is_efficient = np.ones(costs.shape[0], dtype=bool)
    for index, cost in enumerate(costs):
        if is_efficient[index]:
            # Keep any point with a lower cost
            is_efficient[is_efficient] = np.any(costs[is_efficient] < cost,
                                                axis=1)
            # And keep self
            is_efficient[index] = True
    return is_efficient



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""

from typing import List, Dict, Type, Tuple, Union, Any, TextIO, BinaryIO, \
    get_type_hints, get_origin, get_args


from dessia_common.typings import Measure, JsonSerializable,\
    Subclass, InstanceOf, MethodType, ClassMethodType
    
import json
import collections
from importlib import import_module


TYPING_EQUIVALENCES = {int: 'number', float: 'number',
                       bool: 'boolean', str: 'string'}


def full_classname(object_, compute_for: str = 'instance'):
    if compute_for == 'instance':
        return object_.__class__.__module__ + '.' + object_.__class__.__name__
    elif compute_for == 'class':
        return object_.__module__ + '.' + object_.__name__
    else:
        msg = 'Cannot compute {} full classname for object {}'
        raise NotImplementedError(msg.format(compute_for, object_))
        

def is_jsonable(x):
    try:
        json.dumps(x)
        return True
    except:
        return False


def is_sequence(obj):
    """
    :param obj: Object to check
    :return: bool. True if object is a sequence but not a string.
                   False otherwise
    """
    return isinstance(obj, collections.abc.Sequence)\
        and not isinstance(obj, str)


def is_builtin(type_):
    return type_ in TYPING_EQUIVALENCES


def isinstance_base_types(obj):
    return isinstance(obj, str) or isinstance(obj, float) or isinstance(obj, int) or (obj is None)


def get_python_class_from_class_name(full_class_name):
    module_name, class_name = full_class_name.rsplit('.', 1)
    module = import_module(module_name)
    class_ = getattr(module, class_name)
    return class_


def unfold_deep_annotation(typing_=None):
    if is_typing(typing_):
        origin = get_origin(typing_)
        args = get_args(typing_)
        return origin, args
    return None, None


def is_typing(object_: Any):
    has_module = hasattr(object_, '__module__')
    if has_module:
        in_typings = object_.__module__ in ['typing', 'dessia_common.typings']
    else:
        return False
    has_origin = hasattr(object_, '__origin__')
    has_args = hasattr(object_, '__args__')
    return has_module and has_origin and has_args and in_typings


def serialize_typing(typing_):
    if is_typing(typing_):
        origin = get_origin(typing_)
        args = get_args(typing_)
        if origin is Union:
            if len(args) == 2 and type(None) in args:
                # This is a false Union => Is a default value set to None
                return serialize_typing(args[0])
            else:
                # Types union
                argnames = ', '.join([type_fullname(a) for a in args])
                return 'Union[{}]'.format(argnames)
        elif origin is list:
            return 'List[' + type_fullname(args[0]) + ']'
        elif origin is tuple:
            argnames = ', '.join([type_fullname(a) for a in args])
            return 'Tuple[{}]'.format(argnames)
        elif origin is collections.Iterator:
            return 'Iterator[' + type_fullname(args[0]) + ']'
        elif origin is dict:
            key_type = type_fullname(args[0])
            value_type = type_fullname(args[1])
            return 'Dict[{}, {}]'.format(key_type, value_type)
        elif origin is InstanceOf:
            return 'InstanceOf[{}]'.format(type_fullname(args[0]))
        elif origin is Subclass:
            return 'Subclass[{}]'.format(type_fullname(args[0]))
        elif origin is MethodType:
            return 'MethodType[{}]'.format(type_fullname(args[0]))
        elif origin is ClassMethodType:
            return 'ClassMethodType[{}]'.format(type_fullname(args[0]))
        else:
            msg = 'Serialization of typing {} is not implemented'
            raise NotImplementedError(msg.format(typing_))
    if isinstance(typing_, type):
        return full_classname(typing_, compute_for='class')
    if typing_ is TextIO:
        return "TextFile"
    if typing_ is BinaryIO:
        return "BinaryFile"
    return str(typing_)


def type_fullname(arg):
    if arg.__module__ == 'builtins':
        full_argname = '__builtins__.' + arg.__name__
    else:
        full_argname = serialize_typing(arg)
    return full_argname


def type_from_argname(argname):
    splitted_argname = argname.rsplit('.', 1)
    if argname:
        if splitted_argname[0] != '__builtins__':
            return get_python_class_from_class_name(argname)
        # TODO Check for dangerous eval
        return eval(splitted_argname[1])
    return Any


def deserialize_typing(serialized_typing):
    # TODO : handling recursive deserialization
    if isinstance(serialized_typing, str):
        # TODO other builtins should be implemented
        if serialized_typing in ['float', 'builtins.float']:
            return float
        if serialized_typing in ['int', 'builtins.int']:
            return int
        if serialized_typing in ['str', 'builtins.str']:
            return str
        if serialized_typing in ['bool', 'builtins.bool']:
            return bool

        if serialized_typing == "TextFile":
            return TextIO
        if serialized_typing == "BinaryFile":
            return BinaryIO

        if '[' in serialized_typing:
            toptype, remains = serialized_typing.split('[', 1)
            full_argname = remains.rsplit(']', 1)[0]
        else:
            toptype = serialized_typing
            full_argname = ''
        if toptype == 'List':
            return List[type_from_argname(full_argname)]
        elif toptype == 'Tuple':
            if ', ' in full_argname:
                args = full_argname.split(', ')
                if len(args) == 0:
                    return Tuple
                elif len(args) == 1:
                    type_ = type_from_argname(args[0])
                    return Tuple[type_]
                elif len(set(args)) == 1:
                    type_ = type_from_argname(args[0])
                    return Tuple[type_, ...]
                else:
                    msg = ("Heterogenous tuples are forbidden as types for"
                           "workflow non-block variables.")
                    raise TypeError(msg)
            return Tuple[type_from_argname(full_argname)]
        elif toptype == 'Dict':
            args = full_argname.split(', ')
            key_type = type_from_argname(args[0])
            value_type = type_from_argname(args[1])
            return Dict[key_type, value_type]
        # elif splitted_type[0] == 'Union':
        #     args = full_argname.split(', ')
        return get_python_class_from_class_name(serialized_typing)
    raise NotImplementedError('{} of type {}'.format(serialized_typing,
                                                     type(serialized_typing)))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""

# import collections
import tempfile
# import numpy as npy

# import dessia_common.core
from dessia_common.breakdown import object_breakdown


# from openpyxl.writer.excel import save_virtual_workbook
from openpyxl.styles.borders import Border, Side
from openpyxl.styles import Alignment, PatternFill, Font
from openpyxl import Workbook
import openpyxl.utils


def is_hashable(v):
    """Determine whether `v` can be hashed."""
    try:
        hash(v)
    except TypeError:
        return False
    return True


def is_number(v):
    return isinstance(v, int) or isinstance(v, float)


def is_builtins_list(l):
    for e in l:
        if not (is_number(e) or isinstance(e, str)):
            return False
    return True


class XLSXWriter:

    max_column_width = 40
    color_dessIA1 = "263238"
    color_dessIA2 = "537CB0"
    grey1 = "f1f1f1"
    white_font = Font(color="FFFFFF")

    thin_border = Border(left=Side(style='thin'),
                         right=Side(style='thin'),
                         top=Side(style='thin'),
                         bottom=Side(style='thin'))

    def __init__(self, object_):

        self.pattern_color1 = PatternFill(
            fill_type="solid",
            start_color=self.color_dessIA1,
            end_color=self.color_dessIA1)

        self.pattern_color2 = PatternFill(
            fill_type="solid",
            start_color=self.color_dessIA2,
            end_color=self.color_dessIA2)

        self.workbook = Workbook()
        self.main_sheet = self.workbook.active
        self.object = object_

        self.paths = object_breakdown(object_)

        self.classes_to_sheets = {}
        self.object_to_sheet_row = {}
        for class_name, obj_paths in self.paths.items():
            sheet = self.workbook.create_sheet(class_name)
            self.classes_to_sheets[class_name] = sheet

            for i, (obj, path) in enumerate(obj_paths.items()):
                self.object_to_sheet_row[obj] = (sheet, i + 2, path)

        self.write()

    def write_class_header_to_row(self, obj_of_class, sheet, row_number, path=''):
        cell = sheet.cell(row=row_number, column=1, value='Path')
        cell.fill = self.pattern_color2
        cell.border = self.thin_border
        cell.font = self.white_font

        cell = sheet.cell(row=row_number, column=2, value='name')
        cell.fill = self.pattern_color2
        cell.border = self.thin_border
        cell.font = self.white_font
        i = 3

        for (k, _) in sorted(obj_of_class.__dict__.items()):
            if (not k.startswith('_')) and k != 'name':
                cell = sheet.cell(row=row_number, column=i, value=str(k))

                cell.border = self.thin_border
                cell.fill = self.pattern_color2
                cell.font = self.white_font
                i += 1

    def write_object_to_row(self, obj, sheet, row_number, path=''):
        cell = sheet.cell(row=row_number, column=1, value=path)
        cell.border = self.thin_border
        if hasattr(obj, 'name'):
            cell = sheet.cell(row=row_number, column=2, value=obj.name)
        else:
            cell = sheet.cell(row=row_number, column=2, value='No name in model')
        cell.border = self.thin_border
        i = 3
        for (k, v) in sorted(obj.__dict__.items()):
            if (not k.startswith('_')) and k != 'name':
                cell_link = None
                if isinstance(v, dict):
                    str_v = 'Dict of {} items'.format(len(v))
                elif isinstance(v, list):
                    if is_builtins_list(v):
                        str_v = str(v)
                    else:
                        str_v = 'List of {} items'.format(len(v))

                elif isinstance(v, set):
                    str_v = 'Set of {} items'.format(len(v))
                elif isinstance(v, float):
                    str_v = round(v, 6)
                elif is_hashable(v) and v in self.object_to_sheet_row:
                    ref_sheet, ref_row_number, ref_path = self.object_to_sheet_row[v]
                    str_v = ref_path
                    cell_link = '#{}!A{}'.format(ref_sheet.title, ref_row_number)
                else:
                    str_v = str(v)

                cell = sheet.cell(row=row_number, column=i, value=str_v)
                if cell_link:
                    cell.hyperlink = cell_link

                cell.border = self.thin_border

                i += 1

                # column_width = min((len(k) + 1.5), max_column_width)
                # column_name = openpyxl.utils.cell.get_column_letter(i)
                # sheet.column_dimensions[column_name].width = column_width

    def write_object_id(self, sheet):
        sheet.title = 'Object {}'.format(self.object.__class__.__name__)

        sheet['A1'] = 'Module'
        sheet['B1'] = 'Class'
        sheet['C1'] = 'name'

        sheet['A1'].border = self.thin_border
        sheet['B1'].border = self.thin_border
        sheet['C1'].border = self.thin_border
        sheet['A1'].fill = self.pattern_color1
        sheet['B1'].fill = self.pattern_color1
        sheet['C1'].fill = self.pattern_color1
        sheet['A1'].font = self.white_font
        sheet['B1'].font = self.white_font
        sheet['C1'].font = self.white_font

        sheet['A2'] = self.object.__module__
        sheet['B2'] = self.object.__class__.__name__
        sheet['C2'] = self.object.name

        sheet['A2'].border = self.thin_border
        sheet['B2'].border = self.thin_border
        sheet['C2'].border = self.thin_border
        sheet['A2'].fill = self.pattern_color1
        sheet['B2'].fill = self.pattern_color1
        sheet['C2'].fill = self.pattern_color1
        sheet['A2'].font = self.white_font
        sheet['B2'].font = self.white_font
        sheet['C2'].font = self.white_font

        sheet['A3'] = 'Attribute'
        sheet['A4'] = 'Value'
        sheet['A3'].border = self.thin_border
        sheet['A4'].border = self.thin_border

    def write(self):
        # name_column_width = 0
        self.write_object_id(self.main_sheet)
        self.write_class_header_to_row(self.object, self.main_sheet, 3)
        self.write_object_to_row(self.object, self.main_sheet, 4)
        self.autosize_sheet_columns(self.main_sheet, 5, 30)

        for class_name, obj_paths in self.paths.items():
            sheet = self.classes_to_sheets[class_name]

            for obj, path in obj_paths.items():
                _, row_number, path = self.object_to_sheet_row[obj]
                self.write_object_to_row(obj, sheet, row_number, path)
            self.write_class_header_to_row(obj, sheet, 1)

            sheet.auto_filter.ref = "A1:{}{}".format(openpyxl.utils.cell.get_column_letter(sheet.max_column),
                                                     len(obj_paths) + 1)
            self.autosize_sheet_columns(sheet, 5, 30)

    def save_to_file(self, filepath):

        if isinstance(filepath, str):
            real_filepath = filepath
            if not filepath.endswith('.xlsx'):
                real_filepath += '.xlsx'
        else:
            real_filepath = tempfile.NamedTemporaryFile().name

        self.workbook.save(real_filepath)

        if not isinstance(filepath, str):
            with open(real_filepath, 'rb') as file:
                filepath.seek(0)
                filepath.write(file.read())

    def autosize_sheet_columns(self, sheet, min_width=5, max_width=30):
        # Autosize columns
        for col in sheet.columns:
            width = min_width
            column = col[1].column_letter  # Get the column name
            # Since Openpyxl 2.6, the column name is  ".column_letter" as .column became the column number (1-based)
            for cell in col:
                try:  # Necessary to avoid error on empty cells
                    if len(str(cell.value)) > width:
                        width = len(cell.value)
                except:
                    pass
            if width > 0:
                adjusted_width = min((width + 0.5), max_width)
                sheet.column_dimensions[column].width = adjusted_width
